# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import dataclasses
import datetime
import itertools
from collections.abc import Iterator

import sqlalchemy as sa
from nbxmpp import JID
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Select
from sqlalchemy import select
from sqlalchemy import types
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import foreign
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import relationship
from sqlalchemy.orm import remote as sa_remote
from sqlalchemy.sql import expression as expr
from sqlalchemy.types import TypeEngine

from gajim.common import app
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.base import EpochTimestampType
from gajim.common.storage.base import JIDType
from gajim.common.storage.base import JSONType
from gajim.common.storage.base import StrValueMissingType
from gajim.common.storage.base import VALUE_MISSING
from gajim.common.storage.base import ValueMissingT


class Base(DeclarativeBase):
    type_annotation_map: dict[type[Any], TypeEngine[Any]] = {
        float: types.REAL(),
        str: types.TEXT(),
        int: types.INTEGER(),
    }

    def validate(self) -> None:
        pass


class UtilMixin:
    def get_upsert_values(self) -> dict[str, str]:
        values = {
            f.name: getattr(self, f.name)
            for f in dataclasses.fields(self)  # pyright: ignore
        }
        for key, value in list(values.items()):
            if value is VALUE_MISSING:
                values.pop(key)
                continue

            if key not in self.__upsert_cols__:  # pyright: ignore
                values.pop(key)
        return values

    def get_insert_values(self):
        values = {
            f.name: getattr(self, f.name)
            for f in dataclasses.fields(self)  # pyright: ignore
        }
        for key, value in list(values.items()):
            if value is VALUE_MISSING:
                values.pop(key)
                continue

            if key in self.__no_table_cols__:  # pyright: ignore
                values.pop(key)
        return values


class Account(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "account"
    __table_args__ = (Index("idx_account", "jid", unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    jid: Mapped[JID] = mapped_column(JIDType)


class Remote(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "remote"
    __table_args__ = (Index("idx_remote", "jid", unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    jid: Mapped[JID] = mapped_column(JIDType)


class Thread(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "thread"
    __index_cols__ = ["id", "fk_remote_pk", "fk_account_pk"]
    __no_table_cols__ = []
    __table_args__ = (Index("idx_thread", *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE")
    )
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"))

    id: Mapped[str]

    def get_select_stmt(self) -> Select[Any]:
        return select(Thread).where(
            Thread.id == self.id,
            Thread.fk_remote_pk == self.fk_remote_pk,
            Thread.fk_account_pk == self.fk_account_pk,
        )


class Occupant(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "occupant"
    __index_cols__ = ["id", "fk_remote_pk", "fk_account_pk"]
    __upsert_cols__ = ["fk_real_remote_pk", "nickname", "avatar_sha", "updated_at"]
    __no_table_cols__ = [
        "account_",
        "remote_jid_",
        "remote",
        "real_remote_jid_",
        "real_remote",
    ]
    __table_args__ = (Index("idx_occupant", *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)
    remote: Mapped[Remote] = relationship(
        lazy="raise", foreign_keys=fk_remote_pk, viewonly=True, init=False
    )

    id: Mapped[str]

    real_remote_jid_: JID | None | ValueMissingT = dataclasses.field(
        default=VALUE_MISSING, repr=False
    )
    fk_real_remote_pk: Mapped[int | None] = mapped_column(
        StrValueMissingType, ForeignKey("remote.pk"), default=VALUE_MISSING, init=False
    )
    real_remote: Mapped[Remote | None] = relationship(
        lazy="joined",
        foreign_keys=fk_real_remote_pk,
        default=None,
        viewonly=True,
        init=False,
    )

    nickname: Mapped[str | None] = mapped_column(default=None)

    avatar_sha: Mapped[str | None] = mapped_column(
        StrValueMissingType, default=VALUE_MISSING
    )
    blocked: Mapped[bool] = mapped_column(default=False)
    updated_at: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    def validate(self) -> None:
        if self.remote_jid_ is None:  # pyright: ignore
            return
        if not self.remote_jid_.is_bare:
            raise ValueError("Remote JID must be a bare jid")

    def needs_update(self, existing: Occupant) -> bool:
        return existing.updated_at < self.updated_at

    def get_select_stmt(self) -> Select[Any]:
        return select(Occupant).where(
            Occupant.id == self.id,
            Occupant.fk_remote_pk == self.fk_remote_pk,
            Occupant.fk_account_pk == self.fk_account_pk,
        )

    def get_upsert_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {"updated_at": self.updated_at}
        if self.nickname is not None:
            values["nickname"] = self.nickname

        if isinstance(self.fk_real_remote_pk, int):
            values["fk_real_remote_pk"] = sa.func.coalesce(
                Occupant.fk_real_remote_pk, self.fk_real_remote_pk
            )

        if self.avatar_sha is not VALUE_MISSING:
            values["avatar_sha"] = self.avatar_sha

        return values


class OOB(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "oob"

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey("message.pk", ondelete="CASCADE"), init=False
    )
    url: Mapped[str]
    description: Mapped[str | None]


class OpenGraph(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "opengraph"

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey("message.pk", ondelete="CASCADE"), init=False
    )
    about: Mapped[str]
    title: Mapped[str]
    description: Mapped[str | None] = mapped_column(default=None)
    image_type: Mapped[str | None] = mapped_column(default=None)
    image_bytes: Mapped[bytes | None] = mapped_column(default=None)


class Reply(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "reply"

    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey("message.pk", ondelete="CASCADE"), primary_key=True, init=False
    )
    id: Mapped[str]
    to: Mapped[JID | None] = mapped_column(JIDType)


class Call(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "call"

    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey("message.pk", ondelete="CASCADE"), primary_key=True, init=False
    )
    sid: Mapped[str]
    end_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=None
    )
    state: Mapped[int]


class Encryption(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "encryption"
    __index_cols__ = ["key", "trust", "protocol"]
    __no_table_cols__ = []
    __table_args__ = (Index("idx_encryption", "key", "trust", "protocol", unique=True),)

    pk: Mapped[int] = mapped_column(init=False, primary_key=True)
    protocol: Mapped[int]
    key: Mapped[str]
    trust: Mapped[int]

    def get_select_stmt(self) -> Select[Any]:
        return select(Encryption).where(
            Encryption.key == self.key,
            Encryption.trust == self.trust,
            Encryption.protocol == self.protocol,
        )


class SecurityLabel(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "securitylabel"
    __index_cols__ = ["label_hash", "fk_remote_pk", "fk_account_pk"]
    __upsert_cols__ = ["displaymarking", "fgcolor", "bgcolor", "updated_at"]
    __no_table_cols__ = ["account_", "remote_jid_"]
    __table_args__ = (Index("idx_security_label", *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    label_hash: Mapped[str]
    displaymarking: Mapped[str]
    fgcolor: Mapped[str]
    bgcolor: Mapped[str]
    updated_at: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    def needs_update(self, existing: SecurityLabel) -> bool:
        return existing.updated_at < self.updated_at

    def get_select_stmt(self) -> Select[Any]:
        return select(SecurityLabel).where(
            SecurityLabel.label_hash == self.label_hash,
            SecurityLabel.fk_remote_pk == self.fk_remote_pk,
            SecurityLabel.fk_account_pk == self.fk_account_pk,
        )


class MessageError(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "error"
    __index_cols__ = ["message_id", "fk_remote_pk", "fk_account_pk"]
    __no_table_cols__ = ["account_", "remote_jid_"]
    __table_args__ = (Index("idx_error", *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    message_id: Mapped[str]
    by: Mapped[JID | None] = mapped_column(JIDType)
    type: Mapped[str]
    text: Mapped[str | None]
    condition: Mapped[str]
    condition_text: Mapped[str | None]
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)


class Moderation(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "moderation"
    __index_cols__ = ["stanza_id", "fk_remote_pk", "fk_account_pk"]
    __no_table_cols__ = ["account_", "remote_jid_", "occupant", "occupant_"]
    __table_args__ = (Index("idx_moderation", *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey("occupant.pk"), default=None, init=False
    )

    stanza_id: Mapped[str]
    by: Mapped[JID | None] = mapped_column(JIDType)
    reason: Mapped[str | None]
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)


class DisplayedMarker(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "displayed_marker"
    __no_table_cols__ = ["account_", "remote_jid_", "occupant", "occupant_"]

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)
    remote: Mapped[Remote] = relationship(
        lazy="raise", foreign_keys=fk_remote_pk, viewonly=True, init=False
    )

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey("occupant.pk"), default=None, init=False
    )

    id: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    __table_args__ = (
        Index(
            "idx_displayed_marker",
            "id",
            "fk_remote_pk",
            "fk_account_pk",
            unique=True,
            sqlite_where=fk_occupant_pk.is_(None),
        ),
        Index(
            "idx_displayed_marker_gc",
            "id",
            "fk_remote_pk",
            "fk_occupant_pk",
            "fk_account_pk",
            unique=True,
            sqlite_where=fk_occupant_pk.isnot(None),
        ),
        Index(
            "idx_last_displayed_marker",
            "fk_remote_pk",
            "fk_account_pk",
            sa.text("timestamp DESC"),
        ),
        Index(
            "idx_last_displayed_marker_gc",
            "fk_remote_pk",
            "fk_account_pk",
            "fk_occupant_pk",
            sa.text("timestamp DESC"),
        ),
    )


class Receipt(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "receipt"
    __no_table_cols__ = ["account_", "remote_jid_"]
    __table_args__ = (
        Index(
            "idx_receipt",
            "id",
            "fk_remote_pk",
            "fk_account_pk",
            unique=True,
        ),
    )

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    id: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)


class MAMArchiveState(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "mam_archive_state"
    __index_cols__ = ["fk_remote_pk", "fk_account_pk"]
    __upsert_cols__ = [
        "from_stanza_id",
        "from_stanza_ts",
        "to_stanza_id",
        "to_stanza_ts",
    ]
    __no_table_cols__ = ["account_", "remote_jid_"]
    __table_args__ = (Index("idx_mam_archive_state", *__index_cols__, unique=True),)

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    from_stanza_id: Mapped[str | None] = mapped_column(
        StrValueMissingType, default=VALUE_MISSING
    )
    from_stanza_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=VALUE_MISSING
    )
    to_stanza_id: Mapped[str | None] = mapped_column(
        StrValueMissingType, default=VALUE_MISSING
    )
    to_stanza_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=VALUE_MISSING
    )

    def get_select_stmt(self) -> Select[Any]:
        return select(MAMArchiveState).where(
            MAMArchiveState.fk_remote_pk == self.fk_remote_pk,
            MAMArchiveState.fk_account_pk == self.fk_account_pk,
        )

    def needs_update(self, _existing: MAMArchiveState) -> bool:
        return True


class FileTransferSource(MappedAsDataclass, Base, kw_only=True):
    __tablename__ = "ft_source"

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    fk_filetransfer_pk: Mapped[int] = mapped_column(
        ForeignKey("filetransfer.pk", ondelete="CASCADE"), init=False
    )
    type: Mapped[str]

    __mapper_args__ = {
        "polymorphic_identity": "source",
        "polymorphic_on": "type",
    }


class UrlData(FileTransferSource):
    __tablename__ = "ft_source_urldata"

    fk_ft_source_pk: Mapped[int] = mapped_column(
        ForeignKey("ft_source.pk", ondelete="CASCADE"), primary_key=True, init=False
    )
    target: Mapped[str]
    scheme_data: Mapped[dict[str, Any] | None] = mapped_column(JSONType)

    __mapper_args__ = {
        "polymorphic_load": "selectin",
        "polymorphic_identity": "urldata",
    }


class JingleFT(FileTransferSource):
    __tablename__ = "ft_source_jingleft"

    fk_ft_source_pk: Mapped[int] = mapped_column(
        ForeignKey("ft_source.pk", ondelete="CASCADE"), primary_key=True, init=False
    )
    sid: Mapped[str]

    __mapper_args__ = {
        "polymorphic_load": "selectin",
        "polymorphic_identity": "jingleft",
    }


class JinglePub(FileTransferSource):
    __tablename__ = "ft_source_jinglepub"

    fk_ft_source_pk: Mapped[int] = mapped_column(
        ForeignKey("ft_source.pk", ondelete="CASCADE"), primary_key=True, init=False
    )
    id: Mapped[str]

    __mapper_args__ = {
        "polymorphic_load": "selectin",
        "polymorphic_identity": "jinglepub",
    }


class FileTransfer(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "filetransfer"

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)
    fk_message_pk: Mapped[int] = mapped_column(
        ForeignKey("message.pk", ondelete="CASCADE"), init=False
    )

    date: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=None
    )
    desc: Mapped[str | None] = mapped_column(default=None)
    hash: Mapped[str | None] = mapped_column(default=None)
    hash_algo: Mapped[str | None] = mapped_column(default=None)
    height: Mapped[int | None] = mapped_column(default=None)
    length: Mapped[int | None] = mapped_column(default=None)
    media_type: Mapped[str | None] = mapped_column(default=None)
    name: Mapped[str | None] = mapped_column(default=None)
    size: Mapped[int | None] = mapped_column(default=None)
    width: Mapped[int | None] = mapped_column(default=None)

    state: Mapped[int | None]
    path: Mapped[str | None] = mapped_column(default=None, init=False)

    source: Mapped[list[FileTransferSource]] = relationship(
        lazy="selectin",
        default_factory=list,
        cascade="all, delete",
        passive_deletes=True,
    )


class Reaction(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "reaction"
    __upsert_cols__ = ["emojis", "timestamp"]
    __no_table_cols__ = ["account_", "remote_jid_", "occupant", "occupant_"]

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey("occupant.pk"), default=None, init=False
    )

    id: Mapped[str] = mapped_column()
    direction: Mapped[int] = mapped_column()
    emojis: Mapped[str] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)

    __table_args__ = (
        Index(
            "idx_reaction_id",
            "id",
            "fk_remote_pk",
            "fk_account_pk",
            "direction",
            unique=True,
            sqlite_where=fk_occupant_pk.is_(None),
        ),
        Index(
            "idx_reaction_id_gc",
            "id",
            "fk_remote_pk",
            "fk_occupant_pk",
            "fk_account_pk",
            "direction",
            unique=True,
            sqlite_where=fk_occupant_pk.isnot(None),
        ),
    )


class Retraction(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "retraction"
    __no_table_cols__ = ["account_", "remote_jid_", "occupant", "occupant_"]

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    occupant_: Occupant | None = dataclasses.field(repr=False)
    occupant: Mapped[Occupant | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey("occupant.pk"), default=None, init=False
    )

    id: Mapped[str] = mapped_column()
    direction: Mapped[int] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)
    state: Mapped[int] = mapped_column(default=0)

    __table_args__ = (
        Index(
            "idx_retraction_id",
            "id",
            "fk_remote_pk",
            "fk_account_pk",
            "direction",
            unique=True,
            sqlite_where=fk_occupant_pk.is_(None),
        ),
        Index(
            "idx_retraction_id_gc",
            "id",
            "fk_remote_pk",
            "fk_occupant_pk",
            "fk_account_pk",
            "direction",
            unique=True,
            sqlite_where=fk_occupant_pk.isnot(None),
        ),
        Index(
            "idx_retraction_state",
            "state",
        ),
    )


class Message(MappedAsDataclass, Base, UtilMixin, kw_only=True):
    __tablename__ = "message"

    pk: Mapped[int] = mapped_column(primary_key=True, init=False)

    account_: str = dataclasses.field(repr=False)
    account: Mapped[Account] = relationship(lazy="joined", viewonly=True, init=False)
    fk_account_pk: Mapped[int] = mapped_column(
        ForeignKey("account.pk", ondelete="CASCADE"), init=False
    )

    remote_jid_: JID = dataclasses.field(repr=False)
    remote: Mapped[Remote] = relationship(lazy="joined", viewonly=True, init=False)
    fk_remote_pk: Mapped[int] = mapped_column(ForeignKey("remote.pk"), init=False)

    resource: Mapped[str | None] = mapped_column()
    type: Mapped[int] = mapped_column()
    direction: Mapped[int] = mapped_column()
    timestamp: Mapped[datetime.datetime] = mapped_column(EpochTimestampType)
    state: Mapped[int]
    id: Mapped[str | None] = mapped_column(default=None)
    stanza_id: Mapped[str | None] = mapped_column(default=None)
    text: Mapped[str | None] = mapped_column(default=None)
    markup_type: Mapped[int | None] = mapped_column(default=None)
    markup: Mapped[str | None] = mapped_column(default=None)
    user_delay_ts: Mapped[datetime.datetime | None] = mapped_column(
        EpochTimestampType, default=None
    )

    thread_id_: str | None = dataclasses.field(repr=False, default=None)
    thread: Mapped[Thread | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_thread_pk: Mapped[int | None] = mapped_column(
        ForeignKey("thread.pk"), default=None, init=False
    )

    occupant_: Occupant | None = dataclasses.field(repr=False, default=None)
    occupant: Mapped[Occupant | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_occupant_pk: Mapped[int | None] = mapped_column(
        ForeignKey("occupant.pk"), default=None, init=False
    )

    encryption_: Encryption | None = dataclasses.field(repr=False, default=None)
    encryption: Mapped[Encryption | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_encryption_pk: Mapped[int | None] = mapped_column(
        ForeignKey("encryption.pk"), default=None, init=False
    )

    security_label_: SecurityLabel | None = dataclasses.field(repr=False, default=None)
    security_label: Mapped[SecurityLabel | None] = relationship(
        lazy="joined", viewonly=True, init=False
    )
    fk_security_label_pk: Mapped[int | None] = mapped_column(
        ForeignKey("securitylabel.pk"), default=None, init=False
    )

    correction_id: Mapped[str | None] = mapped_column(default=None)

    corrections: Mapped[list[Message]] = relationship(
        lazy="selectin",
        init=False,
        primaryjoin=sa.and_(
            foreign(id) == sa_remote(correction_id),
            foreign(fk_remote_pk) == sa_remote(fk_remote_pk),
            foreign(fk_account_pk) == sa_remote(fk_account_pk),
            foreign(fk_occupant_pk).is_(sa_remote(fk_occupant_pk)),
            foreign(direction) == sa_remote(direction),
            expr.case(
                (
                    sa.and_(
                        foreign(type) == 2,  # Groupchat
                        foreign(fk_occupant_pk).is_(None),
                    ),
                    foreign(resource) == sa_remote(resource),
                ),
                else_=1,
            ),
        ),
        join_depth=1,
        viewonly=True,
        uselist=True,
        order_by=timestamp,
    )

    markers: Mapped[list[DisplayedMarker]] = relationship(
        lazy="raise",
        init=False,
        primaryjoin=sa.and_(
            expr.case(
                (
                    type == 2,  # Groupchat
                    foreign(stanza_id) == DisplayedMarker.id,
                ),
                else_=foreign(id) == DisplayedMarker.id,
            ),
            fk_remote_pk == DisplayedMarker.fk_remote_pk,
            fk_account_pk == DisplayedMarker.fk_account_pk,
        ),
        viewonly=True,
        uselist=True,
        repr=False,
    )

    receipt: Mapped[Receipt | None] = relationship(
        lazy="selectin",
        init=False,
        primaryjoin=sa.and_(
            foreign(id) == Receipt.id,
            fk_remote_pk == Receipt.fk_remote_pk,
            fk_account_pk == Receipt.fk_account_pk,
        ),
        viewonly=True,
    )

    reactions: Mapped[list[Reaction]] = relationship(
        lazy="selectin",
        init=False,
        primaryjoin=sa.and_(
            expr.case(
                (
                    type == 2,  # Groupchat
                    foreign(stanza_id) == Reaction.id,
                ),
                else_=foreign(id) == Reaction.id,
            ),
            fk_remote_pk == Reaction.fk_remote_pk,
            fk_account_pk == Reaction.fk_account_pk,
        ),
        viewonly=True,
        uselist=True,
    )

    retraction: Mapped[Retraction | None] = relationship(
        lazy="joined",
        default=None,
        init=False,
        primaryjoin=sa.and_(
            expr.case(
                (
                    type == 2,  # Groupchat
                    foreign(stanza_id) == Retraction.id,
                ),
                else_=foreign(id) == Retraction.id,
            ),
            fk_remote_pk == Retraction.fk_remote_pk,
            fk_account_pk == Retraction.fk_account_pk,
            fk_occupant_pk.is_(Retraction.fk_occupant_pk),
            direction == Retraction.direction,
        ),
        viewonly=True,
    )

    moderation: Mapped[Moderation | None] = relationship(
        lazy="joined",
        default=None,
        init=False,
        primaryjoin=sa.and_(
            foreign(stanza_id) == sa_remote(Moderation.stanza_id),
            fk_remote_pk == Moderation.fk_remote_pk,
            fk_account_pk == Moderation.fk_account_pk,
        ),
        viewonly=True,
    )

    error: Mapped[MessageError | None] = relationship(
        lazy="joined",
        default=None,
        init=False,
        primaryjoin=sa.and_(
            foreign(id) == sa_remote(MessageError.message_id),
            fk_remote_pk == MessageError.fk_remote_pk,
            fk_account_pk == MessageError.fk_account_pk,
        ),
        viewonly=True,
    )

    filetransfers: Mapped[list[FileTransfer]] = relationship(
        lazy="selectin",
        default_factory=list,
        cascade="all, delete",
        passive_deletes=True,
    )
    call: Mapped[Call | None] = relationship(
        lazy="joined",
        default=None,
        cascade="all, delete",
        passive_deletes=True,
    )
    oob: Mapped[list[OOB]] = relationship(
        lazy="selectin",
        default_factory=list,
        cascade="all, delete",
        passive_deletes=True,
    )
    og: Mapped[list[OpenGraph]] = relationship(
        lazy="selectin",
        default_factory=list,
        cascade="all, delete",
        passive_deletes=True,
    )
    reply: Mapped[Reply | None] = relationship(
        lazy="joined",
        default=None,
        cascade="all, delete",
        passive_deletes=True,
    )

    __table_args__ = (
        Index(
            "idx_message", "fk_remote_pk", "fk_account_pk", sa.text("timestamp DESC")
        ),
        Index(
            "idx_message_corrections",
            "correction_id",
            "fk_remote_pk",
            "fk_account_pk",
            sa.text("timestamp ASC"),
        ),
        Index(
            "idx_message_id",
            "id",
            "fk_remote_pk",
            "fk_account_pk",
        ),
        Index(
            "idx_stanza_id",
            "stanza_id",
            "fk_remote_pk",
            "fk_account_pk",
        ),
    )

    def is_retracted(self) -> bool:
        if self.retraction is None and self.moderation is None:
            return False

        message = self.get_last_correction()
        return message is None

    def get_last_correction(self) -> Message | None:
        for message in reversed(self.corrections):
            if message.retraction is None and message.moderation is None:
                return message

        return None

    def get_referenced_message(self) -> Message | None:
        if self.reply is None:
            return None
        account = app.settings.get_account_from_jid(self.account.jid)
        return app.storage.archive.get_referenced_message(
            account, self.remote.jid, self.type, self.reply.id
        )

    def get_ids_for_moderate(self) -> list[str]:
        stanza_ids: list[str] = []
        for message in itertools.chain([self], self.corrections):
            if message.moderation is None and message.stanza_id is not None:
                stanza_ids.append(message.stanza_id)

        return stanza_ids

    def get_ids_for_retract(self) -> list[str]:
        ids: list[str] = []
        if self.type == MessageType.GROUPCHAT:
            for message in itertools.chain([self], self.corrections):
                if message.retraction is None and message.stanza_id is not None:
                    ids.append(message.stanza_id)
        else:
            for message in itertools.chain([self], self.corrections):
                if message.retraction is None and message.id is not None:
                    ids.append(message.id)

        return ids

    def get_displayed_id(self) -> str | None:
        if self.type == MessageType.GROUPCHAT:
            return self.stanza_id
        return self.id

    def get_reactions(self) -> list[Reaction]:
        # Search in all revisions for reactions but return only the latest
        reactions: dict[int, Reaction] = {}
        if self.type == MessageType.GROUPCHAT:
            for message in itertools.chain([self], self.corrections):
                for reaction in message.reactions:
                    assert reaction.occupant is not None
                    reactions[reaction.occupant.pk] = reaction
        else:
            for message in itertools.chain([self], self.corrections):
                for reaction in message.reactions:
                    reactions[reaction.direction] = reaction

        return list(reactions.values())

    def iter_message_ids(self) -> Iterator[str]:
        for message in itertools.chain([self], self.corrections):
            if message.id is not None:
                yield message.id

    def iter_stanza_ids(self) -> Iterator[str]:
        for message in itertools.chain([self], self.corrections):
            if message.stanza_id is not None:
                yield message.stanza_id



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0080: User Location

from __future__ import annotations

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import LocationData
from nbxmpp.structs import MessageProperties

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.dbus.location import LocationListener
from gajim.common.events import LocationChanged
from gajim.common.events import SignedIn
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import event_node
from gajim.common.modules.util import store_publish
from gajim.common.util.decorators import event_filter


class UserLocation(BaseModule):

    _nbxmpp_extends = 'Location'
    _nbxmpp_methods = [
        'set_location',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._register_pubsub_handler(self._location_received)

        self._current_location: LocationData | None = None
        self._contact_locations: dict[JID, LocationData | None] = {}

    def get_current_location(self) -> LocationData | None:
        return self._current_location

    def get_contact_location(self, jid: JID) -> LocationData | None:
        return self._contact_locations.get(jid)

    @event_node(Namespace.LOCATION)
    def _location_received(self,
                           _con: types.NBXMPPClient,
                           _stanza: Message,
                           properties: MessageProperties
                           ) -> None:
        assert properties.pubsub_event is not None
        if properties.pubsub_event.retracted:
            return

        data = properties.pubsub_event.data
        if properties.is_self_message:
            self._current_location = data

        assert properties.jid is not None
        self._contact_locations[properties.jid] = data

        contact = self._get_contact(properties.jid)
        contact.notify('location-update', data)

    @store_publish
    def set_location(self,
                     location: LocationData | None,
                     force: bool = False
                     ) -> None:
        if not self._con.get_module('PEP').supported:
            return

        if not force and not app.settings.get_account_setting(
                self._account, 'publish_location'):
            return

        if location == self._current_location:
            return

        self._current_location = location
        self._log.info('Send %s', location)
        self._nbxmpp('Location').set_location(location)

    def set_enabled(self, enable: bool) -> None:
        if enable:
            self.register_events([
                ('location-changed', ged.CORE, self._on_location_changed),
                ('signed-in', ged.CORE, self._on_signed_in),
            ])
            self._publish_current_location()
        else:
            self.unregister_events()
            self.set_location(None, force=True)

    def _publish_current_location(self):
        self.set_location(LocationListener.get().current_location)

    @event_filter(['account'])
    def _on_signed_in(self, _event: SignedIn) -> None:
        self._publish_current_location()

    def _on_location_changed(self, event: LocationChanged) -> None:
        if self._current_location == event.info:
            return

        self.set_location(event.info)



# Copyright (C) 2003-2005 Vincent Hanquez <tab AT snarc.org>
# Copyright (C) 2005      Alex Podaras <bigpod AT gmail.com>
#                         Stéphan Kochen <stephan AT kochen.nl>
#                         Alex Mauer <hawke AT hawkesnest.net>
# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
#                         Travis Shirk <travis AT pobox.com>
# Copyright (C) 2005-2008 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006      Junglecow J <junglecow AT gmail.com>
# Copyright (C) 2006-2007 Travis Shirk <travis AT pobox.com>
#                         Stefan Bethge <stefan AT lanpartei.de>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007      James Newton <redshodan AT gmail.com>
#                         Lukas Petrovicky <lukas AT petrovicky.net>
# Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
#                         Julien Pivotto <roidelapluie AT gmail.com>
#                         Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008      Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2018      Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import json
import sys
from dataclasses import dataclass
from pathlib import Path

from packaging.requirements import Requirement
from packaging.version import Version as V

import gajim

from .helpers import is_shipped_plugin
from .plugins_i18n import _ as p_

BLOCKED_PLUGINS = [
    "appindicator_integration",
    "omemo",
    "plugin_installer",
    "syntax_highlight",
    "url_image_preview",
]

GAJIM_VERSION = V(gajim.__version__.split("+", 1)[0]).base_version


@dataclass(frozen=True, eq=True)
class PluginManifest:
    name: str
    short_name: str
    description: str
    authors: list[str]
    homepage: str
    config_dialog: bool
    version: V
    requirements: list[Requirement]
    platforms: list[str]
    path: Path | None = None

    def __hash__(self):
        return hash(f"{self.short_name}_{self.version}")

    @property
    def is_usable(self) -> bool:
        if self.short_name in BLOCKED_PLUGINS:
            return False

        if not self._check_requirements():
            return False

        platform = sys.platform
        if platform not in ("win32", "darwin", "linux"):
            # sys.platform can return an unknown amount of unix/linux derivates
            platform = "others"

        return platform in self.platforms

    @property
    def is_shipped(self) -> bool:
        if self.path is None:
            return False
        return is_shipped_plugin(self.path)

    def get_remote_url(self, repository_url: str) -> str:
        filename = f"{self.short_name}_{self.version}.zip"
        return f"{repository_url}/{self.short_name}/{filename}"

    @classmethod
    def from_path(cls, path: Path) -> PluginManifest:
        manifest_path = path / "plugin-manifest.json"
        if not manifest_path.exists():
            raise ValueError(f"Not a plugin path: {path}")

        if manifest_path.is_dir():
            raise ValueError(f"Not a plugin path: {path}")

        with manifest_path.open(encoding="utf8") as f:
            try:
                manifest = json.load(f)
            except Exception as error:
                raise ValueError(f"Error while parsing manifest: {path}, {error}")
        return cls.from_manifest_json(manifest, manifest_path)

    def _check_requirements(self) -> bool:
        return any(GAJIM_VERSION in req.specifier for req in self.requirements)

    @classmethod
    def from_manifest_json(
        cls, manifest: dict[str, Any], path: Path | None = None
    ) -> PluginManifest:
        return cls(
            name=manifest["name"],
            short_name=manifest["short_name"],
            description=p_(manifest["description"]),
            authors=manifest["authors"],
            homepage=manifest["homepage"],
            config_dialog=manifest["config_dialog"],
            version=V(manifest["version"]),
            requirements=[Requirement(r) for r in manifest["requirements"]],
            platforms=manifest["platforms"],
            path=path.parent if path is not None else path,
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from dataclasses import field

from gajim.common.events import ApplicationEvent

from .manifest import PluginManifest


@dataclass
class PluginAdded(ApplicationEvent):
    name: str = field(init=False, default="plugin-added")
    manifest: PluginManifest


@dataclass
class PluginRemoved(ApplicationEvent):
    name: str = field(init=False, default="plugin-removed")
    manifest: PluginManifest



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0070: Verifying HTTP Requests via XMPP

from __future__ import annotations

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Message
from nbxmpp.structs import IqProperties
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import HttpAuth
from gajim.common.modules.base import BaseModule


class HTTPAuth(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._http_auth,
                          ns=Namespace.HTTP_AUTH,
                          priority=45),
            StanzaHandler(name='iq',
                          callback=self._http_auth,
                          typ='get',
                          ns=Namespace.HTTP_AUTH,
                          priority=45)
        ]

    def _http_auth(self,
                   _con: types.NBXMPPClient,
                   stanza: Iq | Message,
                   properties: IqProperties | MessageProperties
                   ) -> None:
        if properties.http_auth is None:
            return

        self._log.info('Auth request received')
        auto_answer = app.settings.get_account_setting(self._account,
                                                       'http_auth')
        if auto_answer in ('yes', 'no'):
            self.build_http_auth_answer(stanza, auto_answer)
            raise nbxmpp.NodeProcessed

        app.ged.raise_event(
            HttpAuth(client=self._con,
                     data=properties.http_auth,
                     stanza=stanza))
        raise nbxmpp.NodeProcessed

    def build_http_auth_answer(self,
                               stanza: Iq | Message,
                               answer: str
                               ) -> None:
        if answer == 'yes':
            self._log.info('Auth request approved')
            confirm = stanza.getTag('confirm')
            reply = stanza.buildReply('result')
            if stanza.getName() == 'message':
                reply.addChild(node=confirm)
            self._con.connection.send(reply)
        elif answer == 'no':
            self._log.info('Auth request denied')
            err = nbxmpp.Error(stanza, nbxmpp.protocol.ERR_NOT_AUTHORIZED)
            self._con.connection.send(err)



# Copyright (C) 2005-2006 Stéphan Kochen <stephan AT kochen.nl>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2006 Tomasz Melcer <liori AT exroot.org>
# Copyright (C) 2007 Stephan Erb <steve-e AT h3c.de>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# The appearance of the treeview, and parts of the dialog, are controlled by
# AgentBrowser (sub-)classes. Methods that probably should be overridden when
# subclassing are: (look at the docstrings and source for additional info)
# - def cleanup(self) *
# - def _create_treemodel(self) *
# - def _add_actions(self)
# - def _clean_actions(self)
# - def update_actions(self)
# - def default_action(self)
# - def _find_item(self, jid, node)
# - def _add_item(self, jid, node, parent_node, item, force)
# - def _update_info(self, iter_, jid, node, identities, features, data)
# - def _update_error(self, iter_, jid, node)
#
# * Should call the super class for this method.
# All others do not have to call back to the super class (but can if they want
# the functionality).
# There are more methods, of course, but this is a basic set.

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Concatenate
from typing import Generic
from typing import TypeVar

import logging
import types
import weakref
from collections.abc import Callable
from collections.abc import Iterator

from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.modules.pubsub import PubSubSubscription
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoIdentity
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import DiscoItem
from nbxmpp.structs import DiscoItems
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.util.jid import InvalidFormat
from gajim.common.util.jid import parse_jid

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.component_search import ComponentSearch
from gajim.gtk.util.icons import icon_exists
from gajim.gtk.util.window import open_window
from gajim.gtk.window import GajimAppWindow

_T = TypeVar("_T")
_InfoCacheT = tuple[list[DiscoIdentity], list[str], list[Any]]

log = logging.getLogger("gajim.gtk.discovery")

LABELS = {
    1: _("This service has not yet responded with detailed information"),
    2: _(
        "This service could not respond with detailed information.\n"
        "It is most likely a legacy service or broken."
    ),
}

# Dictionary mapping category, type pairs to browser class, image pairs.
# This is a function, so we can call it after the classes are declared.
# For the browser class, None means that the service will only be browsable
# when it advertises disco as it's feature, False means it's never browsable.


def _gen_agent_type_info():
    return {
        # Defaults
        (0, 0): (None, None),
        # XMPP server
        ("server", "im"): (ToplevelAgentBrowser, "jabber"),
        ("services", "jabber"): (ToplevelAgentBrowser, "jabber"),
        ("hierarchy", "branch"): (AgentBrowser, "jabber"),
        # Services
        ("conference", "text"): (MucBrowser, "conference"),
        ("headline", "rss"): (AgentBrowser, "rss"),
        ("headline", "weather"): (False, "weather"),
        ("gateway", "weather"): (False, "weather"),
        ("_jid", "weather"): (False, "weather"),
        ("gateway", "sip"): (False, "sip"),
        ("directory", "user"): (None, "jud"),
        ("pubsub", "generic"): (DiscussionGroupsBrowser, "pubsub"),
        ("pubsub", "service"): (DiscussionGroupsBrowser, "pubsub"),
        ("pubsub", "pep"): (DiscussionGroupsBrowser, "pubsub"),
        ("proxy", "bytestreams"): (None, "bytestreams"),  # Socks5 FT proxy
        ("headline", "newmail"): (ToplevelAgentBrowser, "mail"),
        # Transports
        ("conference", "irc"): (ToplevelAgentBrowser, "irc"),
        ("_jid", "irc"): (False, "irc"),
        ("gateway", "irc"): (False, "irc"),
        ("gateway", "gadu-gadu"): (False, "gadu-gadu"),
        ("_jid", "gadugadu"): (False, "gadu-gadu"),
        ("gateway", "http-ws"): (False, "http-ws"),
        ("gateway", "icq"): (False, "icq"),
        ("_jid", "icq"): (False, "icq"),
        ("gateway", "sms"): (False, "sms"),
        ("_jid", "sms"): (False, "sms"),
        ("gateway", "smtp"): (False, "mail"),
        ("gateway", "mrim"): (False, "mrim"),
        ("_jid", "mrim"): (False, "mrim"),
        ("gateway", "facebook"): (False, "facebook"),
        ("_jid", "facebook"): (False, "facebook"),
        ("gateway", "tv"): (False, "tv"),
        ("gateway", "twitter"): (False, "twitter"),
    }


# Category type to "human-readable" description string
_cat_to_descr: dict[str | tuple[str, str], str] = {
    "other": _("Others"),
    "gateway": _("Transports"),
    "_jid": _("Transports"),
    "conference": _("Group Chat"),
}


class CacheItem(Generic[_T]):
    """
    An object to store cache items and their timeouts
    """

    def __init__(self, value: _T) -> None:
        self.value = value
        self.source: int | None = None

    def __call__(self) -> _T:
        return self.value


class CacheDictionary(Generic[_T]):
    """
    A dictionary that keeps items around for only a specific time.  Lifetime is
    in minutes. Getrefresh specifies whether to refresh when an item is merely
    accessed instead of set as well
    """

    def __init__(self, lifetime: int, getrefresh: bool = True) -> None:
        self.lifetime = lifetime * 1000 * 60
        self.getrefresh = getrefresh
        self.cache: dict[str, CacheItem[_T]] = {}

    def cleanup(self) -> None:
        for key in list(self.cache.keys()):
            item = self.cache[key]
            if item.source:
                GLib.source_remove(item.source)
            del self.cache[key]

    def _expire_timeout(self, key: str) -> bool:
        """
        The timeout has expired, remove the object
        """
        if key in self.cache:
            del self.cache[key]
        return False

    def _refresh_timeout(self, key: str) -> None:
        """
        The object was accessed, refresh the timeout
        """
        item = self.cache[key]
        if item.source:
            GLib.source_remove(item.source)
        if self.lifetime:
            source = GLib.timeout_add_seconds(
                int(self.lifetime / 1000), self._expire_timeout, key
            )
            item.source = source

    def __getitem__(self, key: str) -> _T:
        item = self.cache[key]
        if self.getrefresh:
            self._refresh_timeout(key)
        return item()

    def __setitem__(self, key: str, value: _T) -> None:
        item = CacheItem(value)
        self.cache[key] = item
        self._refresh_timeout(key)

    def __delitem__(self, key: str) -> None:
        item = self.cache[key]
        if item.source:
            GLib.source_remove(item.source)
        del self.cache[key]

    def __contains__(self, key: str) -> bool:
        return key in self.cache


def get_agent_address(jid: str | JID, node: str | None = None) -> str:
    """
    Get an agent's address for displaying in the GUI
    """
    if node:
        return f"{node}@{jid}"
    return str(jid)


class Closure:
    """
    A weak reference to a callback with arguments as an object

    Weak references to methods immediately die, even if the object is still
    alive. Besides a handy way to store a callback, this provides a workaround
    that keeps a reference to the object instead.

    Userargs and removeargs must be tuples.
    """

    def __init__(
        self,
        cb: Any,
        userargs: tuple[Any, ...],
        remove: Callable[..., Any],
        removeargs: tuple[Any, ...],
    ) -> None:
        self.userargs = userargs or ()
        self.remove = remove
        self.removeargs = removeargs or ()
        if isinstance(cb, types.MethodType):
            self.meth_self = weakref.ref(cb.__self__, self._remove)
            self.meth_name = cb.__name__

        elif callable(cb):
            self.meth_self = None
            self.cb: weakref.ReferenceType[Any] = weakref.ref(cb, self._remove)

        else:
            raise TypeError("Object is not callable")

    def _remove(self, ref: Any) -> None:
        self.remove(self, *self.removeargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.meth_self:
            obj = self.meth_self()
            cb = getattr(obj, self.meth_name)
        else:
            cb = self.cb()

        if cb is None:
            return

        args = args + self.userargs
        return cb(*args, **kwargs)


class ServicesCache:
    """
    Class that caches our query results. Each connection will have it's own
    ServiceCache instance
    """

    def __init__(self, account: str) -> None:
        self.account = account
        self._items: CacheDictionary[list[DiscoItem]] = CacheDictionary(
            0, getrefresh=False
        )
        self._info: CacheDictionary[_InfoCacheT] = CacheDictionary(0, getrefresh=False)
        self._cbs: dict[tuple[str, str], list[Any]] = {}

    def cleanup(self) -> None:
        self._items.cleanup()
        self._info.cleanup()

    def _clean_closure(self, cb: Any, type_: str, addr: str):
        # A closure died, clean up
        cbkey = (type_, addr)
        try:
            self._cbs[cbkey].remove(cb)
        except KeyError:
            return
        except ValueError:
            return
        # Clean an empty list
        if not self._cbs[cbkey]:
            del self._cbs[cbkey]

    def get_icon(
        self, identities: list[DiscoIdentity] | None = None, addr: str = ""
    ) -> str:
        """
        Return the icon for an agent
        """
        if identities is None:
            identities = []
        # Grab the first identity with an icon
        for identity in identities:
            try:
                cat, type_ = identity.category, identity.type
                info = _agent_type_info[(cat, type_)]
            except KeyError:
                continue
            service_name = info[1]
            if service_name:
                break
        else:
            # Loop fell through, default to unknown
            service_name = addr.split(".", maxsplit=1)[0]

        # Or load it
        icon_name = f"gajim-agent-{service_name}"
        if icon_exists(icon_name):
            return icon_name
        return "gajim-agent-jabber"

    def get_browser(
        self,
        identities: list[DiscoIdentity] | None = None,
        features: list[str] | None = None,
    ) -> type[AgentBrowser] | None:
        """
        Return the browser class for an agent
        """
        if identities is None:
            identities = []
        if features is None:
            features = []
        # First pass, we try to find a ToplevelAgentBrowser
        for identity in identities:
            try:
                cat, type_ = identity.category, identity.type
                info = _agent_type_info[(cat, type_)]
            except KeyError:
                continue
            browser = info[0]
            if browser is ToplevelAgentBrowser:
                return browser

        # Second pass, we haven't found a ToplevelAgentBrowser
        for identity in identities:
            try:
                cat, type_ = identity.category, identity.type
                info = _agent_type_info[(cat, type_)]
            except KeyError:
                continue
            browser = info[0]
            if browser:
                assert not isinstance(browser, bool)
                return browser
        # Namespace.BROWSE is deprecated, but we check for it anyways.
        # Some services list it in features and respond to
        # Namespace.DISCO_ITEMS anyways.
        # Allow browsing for unknown types as well.
        if (
            (not features and not identities)
            or Namespace.DISCO_ITEMS in features
            or Namespace.BROWSE in features
        ):
            return ToplevelAgentBrowser
        return None

    def get_info(
        self,
        jid: str,
        node: str,
        cb: Callable[
            Concatenate[
                str,
                str,
                list[DiscoIdentity] | None,
                list[str] | None,
                list[Any] | None,
                ...,
            ],
            Any,
        ],
        force: bool = False,
        nofetch: bool = False,
        args: tuple[Any, ...] | None = None,
    ) -> None:
        """
        Get info for an agent
        """

        if args is None:
            args = ()

        addr = get_agent_address(jid, node)
        # Check the cache
        if addr in self._info and not force:
            info = self._info[addr]
            cb(jid, node, *info, *args)
            return

        if nofetch:
            return

        # Create a closure object
        cbkey = ("info", addr)
        cb = Closure(cb, userargs=args, remove=self._clean_closure, removeargs=cbkey)
        # Are we already fetching this?
        if cbkey in self._cbs:
            self._cbs[cbkey].append(cb)
        else:
            self._cbs[cbkey] = [cb]
            client = app.get_client(self.account)
            client.get_module("Discovery").disco_info(
                jid, node, callback=self._disco_info_received
            )

    def get_items(
        self,
        jid: str,
        node: str,
        cb: Callable[Concatenate[str, str, list[DiscoItem] | None, ...], None],
        force: bool = False,
        nofetch: bool = False,
        args: tuple[Any, ...] | None = None,
    ) -> None:
        """
        Get a list of items in an agent
        """
        if args is None:
            args = ()

        addr = get_agent_address(jid, node)
        # Check the cache
        if addr in self._items and not force:
            items = self._items[addr]
            cb(jid, node, items, *args)
            return

        if nofetch:
            return

        # Create a closure object
        cbkey = ("items", addr)
        cb = Closure(cb, userargs=args, remove=self._clean_closure, removeargs=cbkey)
        # Are we already fetching this?
        if cbkey in self._cbs:
            self._cbs[cbkey].append(cb)
        else:
            self._cbs[cbkey] = [cb]
            client = app.get_client(self.account)
            client.get_module("Discovery").disco_items(
                jid, node, callback=self._disco_items_received
            )

    def _disco_info_received(self, task: Any) -> None:
        """
        Callback for when we receive an agent's info
        array is (agent, node, identities, features, data)
        """

        try:
            result = cast(DiscoInfo, task.finish())
        except StanzaError as error:
            self._disco_info_error(error)
            return

        identities = result.identities
        if not identities:
            # Ejabberd doesn't send identities when using admin nodes
            identities = [DiscoIdentity(category="server", type="im", name=result.node)]

        self._on_agent_info(
            str(result.jid),
            result.node or "",
            identities,
            result.features,
            result.dataforms,
        )

    def _disco_info_error(self, error: StanzaError) -> None:
        """
        Callback for when a query fails. Even after the browse and agents
        namespaces
        """
        assert error.jid is not None
        addr = get_agent_address(error.jid)

        # Call callbacks
        cbkey = ("info", addr)
        if cbkey in self._cbs:
            for cb in self._cbs[cbkey]:
                cb(str(error.jid), "", None, None, None)
            # clean_closure may have beaten us to it
            if cbkey in self._cbs:
                del self._cbs[cbkey]

    def _on_agent_info(
        self,
        fjid: str,
        node: str,
        identities: list[DiscoIdentity],
        features: list[str],
        dataforms: list[Any],
    ) -> None:
        addr = get_agent_address(fjid, node)

        # Store in cache
        self._info[addr] = (identities, features, dataforms)

        # Call callbacks
        cbkey = ("info", addr)
        if cbkey in self._cbs:
            for cb in self._cbs[cbkey]:
                cb(fjid, node, identities, features, dataforms)
            # clean_closure may have beaten us to it
            if cbkey in self._cbs:
                del self._cbs[cbkey]

    def _disco_items_received(self, task: Any) -> None:
        """
        Callback for when we receive an agent's items
        array is (agent, node, items)
        """

        try:
            result = cast(DiscoItems, task.finish())
        except StanzaError as error:
            self._disco_items_error(error)
            return

        addr = get_agent_address(result.jid, result.node)

        # Store in cache
        self._items[addr] = result.items

        # Call callbacks
        cbkey = ("items", addr)
        if cbkey in self._cbs:
            for cb in self._cbs[cbkey]:
                cb(str(result.jid), result.node, result.items)
            # clean_closure may have beaten us to it
            if cbkey in self._cbs:
                del self._cbs[cbkey]

    def _disco_items_error(self, error: StanzaError) -> None:
        """
        Callback for when a query fails. Even after the browse and agents
        namespaces
        """
        assert error.jid is not None
        addr = get_agent_address(error.jid)

        # Call callbacks
        cbkey = ("items", addr)
        if cbkey in self._cbs:
            for cb in self._cbs[cbkey]:
                cb(str(error.jid), "", None)
            # clean_closure may have beaten us to it
            if cbkey in self._cbs:
                del self._cbs[cbkey]


class ServiceDiscoveryWindow(GajimAppWindow):
    """
    Class that represents the Services Discovery window
    """

    def __init__(
        self,
        account: str,
        jid: str = "",
        node: str = "",
        address_entry: bool = False,
        parent: ServiceDiscoveryWindow | None = None,
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="ServiceDiscoveryWindow",
            default_width=550,
            default_height=550,
            add_window_padding=True,
            header_bar=True,
        )

        self._account = account
        self.parent = parent
        if not jid:
            jid = app.get_hostname_from_account(account)
            node = ""

        self.jid: str | None = None
        self.browser = None
        self.children: list[ServiceDiscoveryWindow] = []
        self.dying = False
        self.node = ""
        self.reloading = False

        # Check connection
        if not app.account_is_available(account):
            InformationAlertDialog(
                _("You are offline"),
                _(
                    "You are not connected to the server. "
                    "Without a connection, you can not browse available services"
                ),
            )
            raise RuntimeError("You must be connected to browse services")

        # Get a ServicesCache object.
        cache = app.services_cache.get(account)
        if cache is None:
            self.cache = ServicesCache(account)
            app.services_cache[account] = self.cache
        else:
            self.cache = cast(ServicesCache, cache)

        self._ui = get_builder("service_discovery_window.ui")
        self.set_child(self._ui.service_discovery)

        # self.window = self._ui.service_discovery_window
        self.services_treeview = self._ui.services_treeview
        self.model = None
        # This is more reliable than the cursor-changed signal.
        selection = self.services_treeview.get_selection()
        self._connect_after(
            selection, "changed", self._on_services_treeview_selection_changed
        )
        self._connect(
            self.services_treeview,
            "row-activated",
            self._on_services_treeview_row_activated,
        )
        self._connect(
            self._ui.address_comboboxtext,
            "changed",
            self._on_address_comboboxtext_changed,
        )
        self._connect(self._ui.browse_button, "clicked", self._on_go_button_clicked)

        self._connect(self, "close-request", self._on_close_request)

        self.services_scrollwin = self._ui.services_scrollwin
        self.progressbar = self._ui.services_progressbar
        self.banner_header = self._ui.banner_agent_header
        self.banner_subheader = self._ui.banner_agent_subheader
        self.banner_icon = self._ui.banner_agent_icon
        self.style_event_id = 0
        self.action_buttonbox = self._ui.action_buttonbox

        # Address combobox
        self.address_comboboxtext = None
        if address_entry:
            self.address_comboboxtext = self._ui.address_comboboxtext

            self.latest_addresses = app.settings.get("latest_disco_addresses").split()
            if jid in self.latest_addresses:
                self.latest_addresses.remove(jid)
            self.latest_addresses.insert(0, jid)
            if len(self.latest_addresses) > 10:
                self.latest_addresses = self.latest_addresses[0:10]
            for j in self.latest_addresses:
                self.address_comboboxtext.append_text(j)
            child = cast(Gtk.Editable, self.address_comboboxtext.get_child())
            child.set_text(jid)
        else:
            # Don't show it at all if we didn't ask for it
            self._ui.address_box.set_visible(False)

        self._initial_state()
        self.travel(jid, node)

    @property
    def account(self) -> str:
        return self._account

    def _initial_state(self):
        """
        Set some initial state on the window. Separated in a method because it's
        handy to use within browser's cleanup method
        """
        self.progressbar.set_visible(False)
        title_text = _("Service Discovery using account %s") % self.account
        self.set_title(title_text)
        self._set_window_banner_text(_("Service Discovery"))
        self.banner_icon.clear()
        self.banner_icon.set_visible(False)  # Just clearing it doesn't work

    def _set_window_banner_text(self, text: str, text_after: str | None = None) -> None:
        self.banner_header.set_text(text)
        if text_after is not None:
            self.banner_subheader.set_visible(True)
            self.banner_subheader.set_text(text_after)
        else:
            self.banner_subheader.set_visible(False)

    def _cleanup(self) -> None:
        self._destroy()

    def _destroy(self, chain: bool = False) -> None:
        """
        Close the browser. This can optionally close its children and propagate
        to the parent. This should happen on actions like register, or join to
        kill off the entire browser chain
        """
        if self.dying:
            return
        self.dying = True

        # self.browser._get_agent_address() would break when no browser.
        if self.browser:
            self.set_visible(False)
            self.browser.cleanup()
            self.browser = None
        self.close()

        for child in self.children[:]:
            child.parent = None
            if chain:
                child._destroy(chain=chain)
                self.children.remove(child)
        if self.parent:
            if self in self.parent.children:
                self.parent.children.remove(self)
            if chain and not self.parent.children:
                self.parent._destroy(chain=chain)
                self.parent = None
        else:
            self.cache.cleanup()

    def reload(self) -> None:
        if not self.jid:
            return
        self.reloading = True
        self.travel(self.jid, self.node)

    def travel(self, jid: str, node: str) -> None:
        """
        Travel to an agent within the current services window
        """
        if self.browser:
            self.browser.cleanup()
            self.browser = None

        # We need to store these, self.browser is not always available.
        self.jid = jid
        self.node = node
        self.cache.get_info(jid, node, self._travel, force=self.reloading)

    def _travel(
        self,
        jid: str,
        node: str,
        identities: list[DiscoIdentity] | None,
        features: list[str] | None,
        data: list[Any] | None,
    ) -> None:
        """
        Continuation of travel
        """
        if self.dying or jid != self.jid or node != self.node:
            return

        if not identities:
            if not self.address_comboboxtext:
                # We can't travel anywhere else.
                self._destroy()

            if self.parent is not None:
                InformationAlertDialog(
                    _("Service Not Found"),
                    _(
                        "There is no service at the address you entered, or it is "
                        "not responding. Check the address and try again."
                    ),
                    parent=self.parent,
                )
                return

        assert features is not None
        assert data is not None

        klass = self.cache.get_browser(identities, features)
        if not klass:
            InformationAlertDialog(
                _("Service Not Browsable"),
                _("This type of service does not contain any items to browse."),
                parent=self,
            )
            return

        self.browser = klass(self.account, jid, node)
        self.browser.prepare_window(self)
        self.browser.browse(force=self.reloading)
        self.reloading = False

    def open(self, jid: str, node: str) -> None:
        """
        Open an agent. By default, this happens in a new window
        """
        win = open_window(
            "ServiceDiscoveryWindow",
            account=self.account,
            jid=jid,
            node=node,
            parent=self,
        )
        self.children.append(win)

    def _on_close_request(self, _widget: Gtk.ApplicationWindow) -> None:
        self._destroy()

    def _on_address_comboboxtext_changed(self, combo: Gtk.ComboBoxText) -> None:
        if combo.get_active() != -1:
            # User selected one of the entries so do auto-visit
            entry = combo.get_child()
            assert isinstance(entry, Gtk.Entry)
            jid = entry.get_text()
            try:
                jid = parse_jid(jid)
            except InvalidFormat as s:
                InformationAlertDialog(_("Invalid Server Name"), str(s))
                return
            self.travel(jid, "")

    def _on_go_button_clicked(self, _button: Gtk.Button) -> None:
        entry = self._ui.address_comboboxtext.get_child()
        assert isinstance(entry, Gtk.Entry)
        jid = entry.get_text()
        try:
            jid = parse_jid(jid)
        except InvalidFormat as s:
            InformationAlertDialog(_("Invalid Server Name"), str(s), parent=self)
            return
        if jid == self.jid:  # jid has not changed
            return
        if jid in self.latest_addresses:
            self.latest_addresses.remove(jid)
        self.latest_addresses.insert(0, jid)
        if len(self.latest_addresses) > 10:
            self.latest_addresses = self.latest_addresses[0:10]

        assert self.address_comboboxtext is not None
        model = cast(Gtk.ListStore, self.address_comboboxtext.get_model())
        model.clear()

        for j in self.latest_addresses:
            self.address_comboboxtext.append_text(j)
        app.settings.set("latest_disco_addresses", " ".join(self.latest_addresses))
        self.travel(jid, "")

    def _on_services_treeview_row_activated(
        self,
        _treeview: Gtk.TreeView,
        path: Gtk.TreePath,
        col: Gtk.TreeViewColumn | None,
    ) -> None:
        if self.browser:
            self.browser.default_action()

    def _on_services_treeview_selection_changed(
        self, _treeselection: Gtk.TreeSelection
    ) -> None:
        if self.browser:
            self.browser.update_actions()


class AgentBrowser:
    """
    Class that deals with browsing agents and appearance of the browser window.
    This class and subclasses should basically be treated as "part" of the
    ServiceDiscoveryWindow class, but had to be separated because this part is
    dynamic
    """

    def __init__(self, account: str, jid: str, node: str) -> None:
        self.account = account
        self.jid = jid
        self.node = node
        self._total_items = 0
        self.browse_button = None
        # This is for some timeout callbacks
        self.active = False

    def _get_agent_address(self):
        """
        Get the agent's address for displaying in the GUI
        """
        return get_agent_address(self.jid, self.node)

    def _set_initial_title(self):
        """
        Set the initial window title based on agent address
        """
        self.window.set_title(
            _("Browsing %(address)s using account %(account)s")
            % {"address": self._get_agent_address(), "account": self.account}
        )
        self.window._set_window_banner_text(  # pyright: ignore
            self._get_agent_address()
        )

    def _create_treemodel(self):
        """
        Create the treemodel for the services treeview. When subclassing, note
        that the first two columns should ALWAYS be of type string and contain
        the JID and node of the item respectively
        """
        # JID, node, name, address
        self.model = Gtk.ListStore(str, str, str, str)
        self.model.set_sort_column_id(3, Gtk.SortType.ASCENDING)
        self.window.services_treeview.set_model(self.model)
        # Name column
        col = Gtk.TreeViewColumn(_("Name"))
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 2)
        self.window.services_treeview.insert_column(col, -1)
        col.set_resizable(True)
        # Address column
        col = Gtk.TreeViewColumn(_("XMPP Address"))
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 3)
        self.window.services_treeview.insert_column(col, -1)
        col.set_resizable(True)
        self.window.services_treeview.set_headers_visible(True)

    def _clean_treemodel(self):
        self.model.clear()
        for col in self.window.services_treeview.get_columns():
            self.window.services_treeview.remove_column(col)
        self.window.services_treeview.set_headers_visible(False)

    def _add_actions(self):
        """
        Add the action buttons to the buttonbox for actions the browser can
        perform
        """
        self.browse_button = Gtk.Button()
        self.browse_button.connect("clicked", self.on_browse_button_clicked)
        self.window.action_buttonbox.append(self.browse_button)
        self.browse_button.set_label(_("Browse"))
        self.browse_button.set_visible(True)

    def _clean_actions(self):
        """
        Remove the action buttons specific to this browser
        """
        if self.browse_button:
            self.window.action_buttonbox.remove(self.browse_button)
            self.browse_button = None

    def _set_title(
        self,
        _jid: str,
        _node: str,
        identities: list[DiscoIdentity] | None,
        _features: list[str] | None,
        _data: list[Any] | None,
    ) -> None:
        """
        Set the window title based on agent info
        """
        # Set the banner and window title

        if identities is None:
            # Query returned error
            return

        name = ""
        if len(identities) > 1:
            # Check if an identity with server category is present
            for _index, identity in enumerate(identities):
                if identity.category == "server" and identity.name is not None:
                    name = identity.name
                    break
                if identities[0].name is not None:
                    name = identities[0].name

        if name:
            self.window._set_window_banner_text(  # pyright: ignore
                self._get_agent_address(), name
            )

        # Add an icon to the banner.
        icon_name = self.cache.get_icon(identities, addr=self._get_agent_address())
        self.window.banner_icon.set_from_icon_name(icon_name)
        self.window.banner_icon.set_visible(True)

    def _clean_title(self) -> None:
        # Everything done here is done in window._initial_state
        # This is for subclasses.
        pass

    def prepare_window(self, window: ServiceDiscoveryWindow) -> None:
        """
        Prepare the service discovery window. Called when a browser is hooked up
        with a ServiceDiscoveryWindow instance
        """
        self.window = window
        self.cache = window.cache

        self._set_initial_title()
        self._create_treemodel()
        self._add_actions()

        self.update_actions()

        self.active = True
        self.cache.get_info(self.jid, self.node, self._set_title)

    def cleanup(self):
        """
        Cleanup when the window intends to switch browsers
        """
        self.active = False

        self._clean_actions()
        self._clean_treemodel()
        self._clean_title()

        self.window._initial_state()  # pyright: ignore

    def on_browse_button_clicked(self, _button: Gtk.Button | None = None) -> None:
        """
        When we want to browse an agent: open a new services window with a
        browser for the agent type
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        jid = model[iter_][0]
        if jid:
            node = model[iter_][1]
            self.window.open(jid, node)

    def update_actions(self) -> None:
        """
        When we select a row: activate action buttons based on the agent's info
        """
        if self.browse_button:
            self.browse_button.set_sensitive(False)
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        jid = model[iter_][0]
        node = model[iter_][1]
        if jid:
            self.cache.get_info(jid, node, self._update_actions, nofetch=True)

    def _update_actions(
        self,
        _jid: str,
        _node: str,
        identities: list[DiscoIdentity] | None,
        features: list[str] | None,
        _data: list[Any] | None,
    ) -> None:
        """
        Continuation of update_actions
        """
        if not identities or not self.browse_button:
            return

        assert features is not None

        klass = self.cache.get_browser(identities, features)
        if klass:
            self.browse_button.set_sensitive(True)

    def default_action(self) -> None:
        """
        When we double-click a row: perform the default action on the selected
        item
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        jid = model[iter_][0]
        node = model[iter_][1]
        if jid:
            self.cache.get_info(jid, node, self._default_action, nofetch=True)

    def _default_action(
        self,
        _jid: str,
        _node: str,
        identities: list[DiscoIdentity] | None,
        features: list[str] | None,
        _data: list[Any] | None,
    ) -> bool:
        """
        Continuation of default_action
        """

        if identities is None:
            # Query returned error
            return False

        assert features is not None

        if self.cache.get_browser(identities, features):
            # Browse if we can
            self.on_browse_button_clicked()
            return True
        return False

    def browse(self, force: bool = False) -> None:
        """
        Fill the treeview with agents, fetching the info if necessary
        """
        self.model.clear()
        self._total_items = self._progress = 0
        self.window.progressbar.set_visible(True)
        self._pulse_timeout = GLib.timeout_add(250, self._pulse_timeout_cb)
        self.cache.get_items(
            self.jid, self.node, self._agent_items, force=force, args=(force,)
        )

    def _pulse_timeout_cb(self, *args: Any) -> bool:
        """
        Simple callback to keep the progressbar pulsing
        """
        if not self.active:
            return False
        self.window.progressbar.pulse()
        return True

    def _find_item(self, jid: str, node: str) -> Gtk.TreeIter | None:
        """
        Check if an item is already in the treeview. Return an iter to it if so,
        None otherwise
        """
        iter_ = self.model.get_iter_first()
        while iter_:
            cjid = self.model.get_value(iter_, 0)
            cnode = self.model.get_value(iter_, 1)
            if jid == cjid and node == cnode:
                break
            iter_ = self.model.iter_next(iter_)
        if iter_:
            return iter_
        return None

    def add_self_line(self) -> None:
        pass

    def _agent_items(
        self, jid: str, node: str, items: list[DiscoItem] | None, force: bool
    ) -> None:
        """
        Callback for when we receive a list of agent items
        """
        self.model.clear()
        self.add_self_line()
        self._total_items = 0
        GLib.source_remove(self._pulse_timeout)
        self.window.progressbar.set_visible(False)
        # The server returned an error
        if not items:
            if self.window.parent is not None:
                InformationAlertDialog(
                    _("Service Not Browsable"),
                    _("This service does not contain any items to browse."),
                    parent=self.window.parent,
                )
            if not self.window.address_comboboxtext:
                # We can't travel anywhere else.
                self.window.close()
            return

        # We got a list of items
        def fill_partial_rows(items: list[DiscoItem]) -> Iterator[bool]:
            """Generator to fill the listmodel of a treeview progressively."""
            for item in items:
                if self.window.dying:
                    yield False

                # If such an item is already here: don't add it
                if self._find_item(str(item.jid), item.node or ""):
                    continue

                self._total_items += 1
                self._add_item(node, item, force)
                if (self._total_items % 10) == 0:
                    yield True
            # Stop idle_add()
            yield False

        loader = fill_partial_rows(items)
        GLib.idle_add(next, loader)

    def _agent_info(
        self,
        jid: str,
        node: str,
        identities: list[DiscoIdentity] | None,
        features: list[str] | None,
        data: list[Any] | None,
    ) -> None:
        """
        Callback for when we receive info about an agent's item
        """

        iter_ = self._find_item(jid, node)
        if not iter_:
            # Not in the treeview, stop
            return

        if identities is None:
            # The server returned an error
            self._update_error(iter_, jid, node)

        else:
            assert features is not None
            assert data is not None
            # We got our info
            self._update_info(iter_, jid, node, identities, features, data)

        self.update_actions()

    def _add_item(self, parent_node: str, item: DiscoItem, force: bool) -> None:
        """
        Called when an item should be added to the model. The result of a
        disco#items query
        """
        node = item.node or ""
        name = item.name or ""
        jid = str(item.jid)

        self.model.append((jid, node, name, get_agent_address(jid, node)))
        self.cache.get_info(jid, node, self._agent_info, force=force)

    def _update_info(
        self,
        iter_: Gtk.TreeIter,
        jid: str,
        node: str,
        identities: list[DiscoIdentity],
        features: list[str],
        data: list[Any],
    ) -> None:
        """
        Called when an item should be updated in the model with further info.
        The result of a disco#info query
        """
        name = identities[0].name or ""
        if name:
            self.model[iter_][2] = name

    def _update_error(self, iter_: Gtk.TreeIter, jid: str, node: str) -> None:
        """Called when a disco#info query failed for an item."""


class ToplevelAgentBrowser(AgentBrowser):
    """
    This browser is used at the top level of a jabber server to browse services
    such as transports, conference servers, etc
    """

    def __init__(self, *args: Any):
        AgentBrowser.__init__(self, *args)
        self._progressbar_sourceid = None
        self._progress = 0
        self.register_button = None
        self.join_button = None
        self.execute_button = None
        self.search_button = None
        self._scroll_signal = None

    def add_self_line(self):
        addr = get_agent_address(self.jid, self.node)
        descr = f"<b>{addr}</b>"
        # Guess which kind of service this is
        identities: list[DiscoIdentity] = []
        type_ = app.get_transport_name_from_jid(self.jid, use_config_setting=False)
        if type_:
            identity = DiscoIdentity(category="_jid", type=type_, name=None)
            identities.append(identity)
        # Set the pixmap for the row
        icon_name = self.cache.get_icon(identities, addr=addr)
        self.model.append(None, [self.jid, self.node, icon_name, descr, LABELS[1]])
        # Grab info on the service
        self.cache.get_info(self.jid, self.node, self._agent_info, force=False)

    def _pixbuf_renderer_data_func(
        self,
        col: Gtk.TreeViewColumn,
        cell: Gtk.CellRenderer,
        model: Gtk.TreeModel,
        iter_: Gtk.TreeIter,
        data: Any = None,
    ) -> None:
        """
        Callback for setting the pixbuf renderer's properties
        """
        jid = model.get_value(iter_, 0)
        if jid:
            icon_name = model.get_value(iter_, 2)
            cell.set_property("icon-size", Gtk.IconSize.LARGE)
            cell.set_property("visible", True)
            cell.set_property("icon_name", icon_name)
        else:
            cell.set_property("visible", False)

    def _text_renderer_data_func(
        self,
        col: Gtk.TreeViewColumn,
        cell: Gtk.CellRenderer,
        model: Gtk.TreeModel,
        iter_: Gtk.TreeIter,
        data: Any = None,
    ) -> None:
        """
        Callback for setting the text renderer's properties
        """
        jid = model.get_value(iter_, 0)
        markup = model.get_value(iter_, 3)
        state = model.get_value(iter_, 4)
        cell.set_property("markup", markup)
        if jid:
            if state is not None:
                # fetching or error
                cell.set_property("foreground_set", True)
            else:
                # Normal/success
                cell.set_property("foreground_set", False)
        else:
            cell.set_property("foreground_set", False)

    def _treemodel_sort_func(
        self,
        model: Gtk.TreeModel,
        iter1: Gtk.TreeIter,
        iter2: Gtk.TreeIter,
        data: Any = None,
    ) -> int:
        """
        Sort function for our treemode
        """
        # Compare state
        state1 = model.get_value(iter1, 4)
        state2 = model.get_value(iter2, 4)
        if state1 is not None:
            return 1
        if state2 is not None:
            return -1
        descr1 = model.get_value(iter1, 3)
        descr2 = model.get_value(iter2, 3)
        # Compare strings
        if descr1 > descr2:
            return 1
        if descr1 < descr2:
            return -1
        return 0

    def _create_treemodel(self) -> None:
        # JID, node, icon, description, state
        # state is None on success or has a string
        # from LABELS on error or while fetching
        view = self.window.services_treeview
        self.model = Gtk.TreeStore(str, str, str, str, str)
        self.model.set_sort_func(4, self._treemodel_sort_func)
        self.model.set_sort_column_id(4, Gtk.SortType.ASCENDING)
        view.set_model(self.model)

        col = Gtk.TreeViewColumn()
        # Icon Renderer
        renderer = Gtk.CellRendererPixbuf()
        renderer.set_property("xpad", 6)
        col.pack_start(renderer, False)
        col.set_cell_data_func(renderer, self._pixbuf_renderer_data_func)
        # Text Renderer
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.set_cell_data_func(renderer, self._text_renderer_data_func)
        renderer.set_property("foreground", "dark gray")

        view.set_tooltip_column(4)
        view.insert_column(col, -1)
        col.set_resizable(True)

    def _clean_treemodel(self) -> None:
        # Disconnect signals
        if self._scroll_signal:
            scrollwin = self.window.services_scrollwin
            scrollwin.disconnect(self._scroll_signal)
            self._scroll_signal = None
        AgentBrowser._clean_treemodel(self)

    def _add_actions(self) -> None:
        AgentBrowser._add_actions(self)
        self.execute_button = Gtk.Button(label=_("_Command"), use_underline=True)
        self.execute_button.connect("clicked", self._on_execute_button_clicked)
        image = Gtk.Image.new_from_icon_name("lucide-terminal-symbolic")
        self.execute_button.set_child(image)
        self.window.action_buttonbox.append(self.execute_button)
        self.execute_button.set_visible(True)

        self.register_button = Gtk.Button(label=_("Re_gister"), use_underline=True)
        self.register_button.connect("clicked", self._on_register_button_clicked)
        self.window.action_buttonbox.append(self.register_button)
        self.register_button.set_visible(True)

        self.join_button = Gtk.Button(label=_("Join"), use_underline=True)
        self.join_button.connect("clicked", self._on_join_button_clicked)
        self.window.action_buttonbox.append(self.join_button)
        self.join_button.set_visible(True)

        self.search_button = Gtk.Button(label=_("_Search"), use_underline=True)
        self.search_button.connect("clicked", self.on_search_button_clicked)
        image = Gtk.Image.new_from_icon_name("lucide-search-symbolic")
        self.search_button.set_child(image)
        self.window.action_buttonbox.append(self.search_button)
        self.search_button.set_visible(True)

    def _clean_actions(self) -> None:
        if self.execute_button:
            self.window.action_buttonbox.remove(self.execute_button)
            self.execute_button = None
        if self.register_button:
            self.window.action_buttonbox.remove(self.register_button)
            self.register_button = None
        if self.join_button:
            self.window.action_buttonbox.remove(self.join_button)
            self.join_button = None
        if self.search_button:
            self.window.action_buttonbox.remove(self.search_button)
            self.search_button = None
        AgentBrowser._clean_actions(self)

    def on_search_button_clicked(self, _button: Gtk.Button | None = None) -> None:
        """
        When we want to search something: open search window
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        service = model[iter_][0]
        ComponentSearch(self.account, service, self.window)

    def cleanup(self) -> None:
        AgentBrowser.cleanup(self)

    def _on_execute_button_clicked(self, _button: Gtk.Button | None = None) -> None:
        """
        When we want to execute a command: open adhoc command window
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        service = model[iter_][0]
        open_window("AdHocCommands", account=self.account, jids=[service])

    def _on_register_button_clicked(self, _button: Gtk.Button | None = None) -> None:
        """
        When we want to register an agent: request information about registering
        with the agent and close the window
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        jid = JID.from_string(model[iter_][0])
        if jid:
            open_window("ServiceRegistration", account=self.account, address=jid)

    def _on_join_button_clicked(self, _button: Gtk.Button | None = None) -> None:
        """
        When we want to join an IRC room or create a new MUC room: Opens the
        join_groupchat_window
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        service = model[iter_][0]
        app.window.show_add_join_groupchat(self.account, service)

    def update_actions(self) -> None:
        if self.execute_button:
            self.execute_button.set_sensitive(False)
        if self.register_button:
            self.register_button.set_sensitive(False)
        if self.browse_button:
            self.browse_button.set_sensitive(False)
        if self.join_button:
            self.join_button.set_sensitive(False)
        if self.search_button:
            self.search_button.set_sensitive(False)
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        if not model[iter_][0]:
            # We're on a category row
            return
        if model[iter_][4] is not None:
            # We don't have the info (yet)
            # It's either unknown or a transport:
            # Register button should be active
            if self.register_button:
                self.register_button.set_sensitive(True)
            # Guess what kind of service we're dealing with
            if self.browse_button:
                jid = model[iter_][0]
                type_ = app.get_transport_name_from_jid(jid, use_config_setting=False)
                if type_:
                    identity = DiscoIdentity(category="_jid", type=type_)
                    klass = self.cache.get_browser([identity])
                    if klass:
                        self.browse_button.set_sensitive(True)
                else:
                    # We couldn't guess
                    self.browse_button.set_sensitive(True)
        else:
            # Normal case, we have info
            AgentBrowser.update_actions(self)

    def _update_actions(
        self,
        jid: str,
        node: str,
        identities: list[DiscoIdentity] | None,
        features: list[str] | None,
        data: list[Any] | None,
    ):
        AgentBrowser._update_actions(self, jid, node, identities, features, data)

        if identities is None:
            # Discoinfo ran into an error
            return

        assert features is not None

        if self.execute_button and Namespace.COMMANDS in features:
            self.execute_button.set_sensitive(True)
        if self.search_button and Namespace.SEARCH in features:
            self.search_button.set_sensitive(True)
        # Don't authorize to register with a server via disco
        if self.register_button and Namespace.REGISTER in features and jid != self.jid:
            # We can register this agent
            registered_transports: list[JID | str] = []
            client = app.get_client(self.account)
            for contact in client.get_module("Roster").iter_contacts():
                if contact.is_gateway:
                    registered_transports.append(contact.jid)
            registered_transports.append(self.jid)
            if jid in registered_transports:
                self.register_button.set_label(_("_Edit"))
            else:
                self.register_button.set_label(_("Re_gister"))
            self.register_button.set_sensitive(True)

    def _default_action(
        self,
        jid: str,
        node: str,
        identities: list[DiscoIdentity] | None,
        features: list[str] | None,
        data: list[Any] | None,
    ) -> bool:
        if AgentBrowser._default_action(self, jid, node, identities, features, data):
            return True

        assert features is not None
        if Namespace.REGISTER in features:
            # Register if we can't browse
            self._on_register_button_clicked()
            return True
        return False

    def browse(self, force: bool = False) -> None:
        self._progress = 0
        AgentBrowser.browse(self, force=force)

    def _expand_all(self) -> None:
        """
        Expand all items in the treeview
        """
        self.window.services_treeview.expand_all()

    def _update_progressbar(self) -> None:
        """
        Update the progressbar
        """
        # Refresh this every update
        if self._progressbar_sourceid:
            GLib.source_remove(self._progressbar_sourceid)

        fraction = 0
        if self._total_items:
            self.window.progressbar.set_text(
                _("Scanning %(current)d / %(total)d ...")
                % {"current": self._progress, "total": self._total_items + 1}
            )
            fraction = float(self._progress) / float(self._total_items)
            if self._progress >= self._total_items:
                # We show the progressbar for just a bit before hiding it.
                id_ = GLib.timeout_add_seconds(2, self._hide_progressbar_cb)
                self._progressbar_sourceid = id_
            else:
                self.window.progressbar.set_visible(True)
                # Hide the progressbar if we're timing out anyways. (20 secs)
                id_ = GLib.timeout_add_seconds(20, self._hide_progressbar_cb)
                self._progressbar_sourceid = id_
        self.window.progressbar.set_fraction(fraction)

    def _hide_progressbar_cb(self, *args: Any) -> bool:
        """
        Simple callback to hide the progressbar a second after we finish
        """
        if self.active:
            self.window.progressbar.set_visible(False)
        return False

    def _friendly_category(self, category: str, type_: str = "") -> str:
        """
        Get the friendly category name
        """
        cat = None
        if type_:
            # Try type-specific override
            try:
                cat = _cat_to_descr[(category, type_)]
            except KeyError:
                pass
        if not cat:
            try:
                cat = _cat_to_descr[category]
            except KeyError:
                cat = _cat_to_descr["other"]
        return cat

    def _create_category(self, cat: str, type_: str = "") -> Gtk.TreeIter:
        """
        Creates a category row
        """
        cat = self._friendly_category(cat, type_)
        return self.model.append(None, ["", "", None, cat, None])

    def _find_category(self, cat: str, type_: str = "") -> Gtk.TreeIter | None:
        """
        Looks up a category row and returns the iterator to it, or None
        """
        cat = self._friendly_category(cat, type_)
        iter_ = self.model.get_iter_first()
        while iter_:
            if self.model.get_value(iter_, 3) == cat:
                break
            iter_ = self.model.iter_next(iter_)
        if iter_:
            return iter_
        return None

    def _find_item(self, jid: str, node: str) -> Gtk.TreeIter | None:
        iter_ = None
        cat_iter = self.model.get_iter_first()
        while cat_iter and not iter_:
            cjid = self.model.get_value(cat_iter, 0)
            cnode = self.model.get_value(cat_iter, 1)
            if jid == cjid and node == cnode:
                iter_ = cat_iter
                break
            iter_ = self.model.iter_children(cat_iter)
            while iter_:
                cjid = self.model.get_value(iter_, 0)
                cnode = self.model.get_value(iter_, 1)
                if jid == cjid and node == cnode:
                    break
                iter_ = self.model.iter_next(iter_)
            cat_iter = self.model.iter_next(cat_iter)
        if iter_:
            return iter_
        return None

    def _add_item(self, parent_node: str, item: DiscoItem, force: bool) -> None:
        node = item.node or ""
        jid = str(item.jid)

        # Row text
        addr = get_agent_address(jid, node)
        if item.name:
            descr = f"<b>{item.name}</b>\n{addr}"
        else:
            descr = f"<b>{addr}</b>"
        # Guess which kind of service this is
        identities: list[DiscoIdentity] = []
        type_ = app.get_transport_name_from_jid(jid, use_config_setting=False)
        if type_:
            identity = DiscoIdentity(category="_jid", type=type_)
            identities.append(identity)
            cat_args = ("_jid", type_)
        else:
            # Put it in the 'other' category for now
            cat_args = ("other",)

        icon_name = self.cache.get_icon(identities, addr=addr)
        # Put it in the right category
        cat = self._find_category(*cat_args)
        if not cat:
            cat = self._create_category(*cat_args)
        self.model.append(cat, [jid, node, icon_name, descr, LABELS[1]])
        GLib.idle_add(self._expand_all)
        # Grab info on the service
        self.cache.get_info(jid, node, self._agent_info, force=force)
        self._update_progressbar()

    def _update_info(
        self,
        iter_: Gtk.TreeIter,
        jid: str,
        node: str,
        identities: list[DiscoIdentity],
        features: list[str],
        data: list[Any],
    ) -> None:
        addr = get_agent_address(jid, node)
        if not identities:
            descr = "<b>{addr}</b>"
        else:
            name = identities[0].name or ""
            if name:
                descr = f"<b>{name}</b>\n{addr}"
            else:
                descr = f"<b>{addr}</b>"

        # Update progress
        self._progress += 1
        self._update_progressbar()

        # Search for an icon and category we can display
        icon_name = self.cache.get_icon(identities, addr=addr)
        cat, type_ = "", ""
        for identity in identities:
            try:
                cat, type_ = identity.category, identity.type
            except KeyError:
                continue
            break

        # Check if we have to move categories
        old_cat_iter = self.model.iter_parent(iter_)
        if not old_cat_iter or self.model.get_value(old_cat_iter, 3) == cat:
            # Already in the right category, just update
            self.model[iter_][2] = icon_name
            self.model[iter_][3] = descr
            self.model[iter_][4] = None
            return
        # Not in the right category, move it.
        self.model.remove(iter_)

        old_cat = self.model.get_value(old_cat_iter, 3)
        # Check if the old category is empty
        if not self.model.iter_is_valid(old_cat_iter):
            old_cat_iter = self._find_category(old_cat)
        if not self.model.iter_children(old_cat_iter):
            assert old_cat_iter is not None
            self.model.remove(old_cat_iter)

        cat_iter = self._find_category(cat, type_)
        if not cat_iter:
            cat_iter = self._create_category(cat, type_)
        self.model.append(cat_iter, [jid, node, icon_name, descr, None])
        self._expand_all()

    def _update_error(self, iter_: Gtk.TreeIter, jid: str, node: str) -> None:
        self.model[iter_][4] = LABELS[2]
        self._progress += 1
        self._update_progressbar()


class MucBrowser(AgentBrowser):
    def __init__(self, *args: Any, **kwargs: Any):
        AgentBrowser.__init__(self, *args, **kwargs)
        self.join_button = None

    def _create_treemodel(self):
        # JID, node, name, users_int, users_str, description, fetched
        # This is rather long, I'd rather not use a data_func here though.
        # Users is a string, because want to be able to leave it empty.
        self.model = Gtk.ListStore(str, str, str, int, str, str, bool)
        self.model.set_sort_column_id(2, Gtk.SortType.ASCENDING)
        self.window.services_treeview.set_model(self.model)
        # Name column
        col = Gtk.TreeViewColumn(_("Name"))
        col.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        col.set_fixed_width(100)
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 2)
        col.set_sort_column_id(2)
        self.window.services_treeview.insert_column(col, -1)
        col.set_resizable(True)
        # Users column
        col = Gtk.TreeViewColumn(_("Users"))
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 4)
        col.set_sort_column_id(3)
        self.window.services_treeview.insert_column(col, -1)
        col.set_resizable(True)
        # Description column
        col = Gtk.TreeViewColumn(_("Description"))
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 5)
        col.set_sort_column_id(4)
        self.window.services_treeview.insert_column(col, -1)
        col.set_resizable(True)
        # Id column
        col = Gtk.TreeViewColumn(_("Id"))
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 0)
        col.set_sort_column_id(0)
        self.window.services_treeview.insert_column(col, -1)
        col.set_resizable(True)
        self.window.services_treeview.set_headers_visible(True)
        self.window.services_treeview.set_headers_clickable(True)
        # Source id for idle callback used to start disco#info queries.
        self._fetch_source = None
        # Query failure counter
        self._broken = 0
        # Connect to scrollwindow scrolling
        self.vadj = self.window.services_scrollwin.get_property("vadjustment")
        self.vadj_cbid = self.vadj.connect("value-changed", self.on_scroll)

    def _clean_treemodel(self) -> None:
        if self.vadj_cbid:
            self.vadj.disconnect(self.vadj_cbid)
            self.vadj_cbid = None
        AgentBrowser._clean_treemodel(self)

    def _add_actions(self) -> None:
        self.join_button = Gtk.Button(label=_("_Join"), use_underline=True)
        self.join_button.connect("clicked", self._on_join_button_clicked)
        self.window.action_buttonbox.append(self.join_button)
        self.join_button.set_visible(True)

    def _clean_actions(self):
        if self.join_button:
            self.window.action_buttonbox.remove(self.join_button)
            self.join_button = None

    def _on_join_button_clicked(self, *args: Any) -> None:
        """
        When we want to join a conference: ask specific information about the
        selected agent and close the window
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_:
            return
        service = model[iter_][0]
        app.window.show_add_join_groupchat(self.account, service)

    def update_actions(self) -> None:
        sens = self.window.services_treeview.get_selection().count_selected_rows()
        if self.join_button:
            self.join_button.set_sensitive(sens > 0)

    def default_action(self) -> None:
        self._on_join_button_clicked()

    def _start_info_query(self) -> bool:
        """
        Idle callback to start checking for visible rows
        """
        self._fetch_source = None
        self._query_visible()
        return False

    def on_scroll(self, *args: Any) -> None:
        """
        Scrollwindow callback to trigger new queries on scrolling
        """
        # This apparently happens when inactive sometimes
        self._query_visible()

    def _query_visible(self) -> None:
        """
        Query the next visible row for info
        """
        if self._fetch_source:
            # We're already fetching
            return
        view = self.window.services_treeview
        if not view.get_realized():
            # Prevent a silly warning, try again in a bit.
            self._fetch_source = GLib.timeout_add(100, self._start_info_query)
            return
        range_ = view.get_visible_range()
        if not range_:
            return
        start, end = range_
        iter_ = self.model.get_iter(start)
        while iter_:
            if not self.model.get_value(iter_, 6):
                jid = self.model.get_value(iter_, 0)
                node = self.model.get_value(iter_, 1)
                self.cache.get_info(jid, node, self._agent_info)
                self._fetch_source = True
                return
            if self.model.get_path(iter_) == end:
                break
            iter_ = self.model.iter_next(iter_)
        self._fetch_source = None

    def _channel_altinfo(
        self,
        jid: str,
        node: str,
        items: list[DiscoItem] | None,
        name: str | None = None,
    ) -> None:
        """
        Callback for the alternate disco#items query. We try to at least get
        the amount of users in the room if the service does not support MUC
        dataforms
        """
        if items is None:
            # The server returned an error
            self._broken += 1
            if self._broken >= 3:
                # Disable queries completely after 3 failures
                if self.vadj_cbid:
                    self.vadj.disconnect(self.vadj_cbid)
                    self.vadj_cbid = None
                self._fetch_source = None
                return
        else:
            iter_ = self._find_item(jid, node)
            if iter_:
                if name:
                    self.model[iter_][2] = name
                self.model[iter_][3] = len(items)  # The number of users
                self.model[iter_][4] = str(len(items))  # The number of users
                self.model[iter_][6] = True
        self._fetch_source = None
        self._query_visible()

    def _add_item(self, parent_node: str | None, item: DiscoItem, force: bool) -> None:
        node = item.node or ""
        name = item.name or ""
        jid = str(item.jid)

        self.model.append((jid, node, name, -1, "", "", False))
        if not self._fetch_source:
            self._fetch_source = GLib.idle_add(self._start_info_query)

    def _update_info(
        self,
        iter_: Gtk.TreeIter,
        jid: str,
        node: str,
        identities: list[DiscoIdentity],
        features: list[str],
        data: list[Any],
    ) -> None:
        name = identities[0].name or ""
        for form in data:
            typefield = form.vars.get("FORM_TYPE")
            if (
                typefield
                and typefield.value == "http://jabber.org/protocol/muc#roominfo"
            ):
                # Fill model row from the form's fields
                users = form.vars.get("muc#roominfo_occupants")
                descr = form.vars.get("muc#roominfo_description")
                if users:
                    self.model[iter_][3] = int(users.value)
                    self.model[iter_][4] = users.value
                if descr and descr.value:
                    self.model[iter_][5] = descr.value
                # Only set these when we find a form with additional info
                # Some servers don't support forms and put extra info in
                # the name attribute, so we preserve it in that case.
                self.model[iter_][2] = name
                self.model[iter_][6] = True
                break
        else:
            # We didn't find a form, switch to alternate query mode
            self.cache.get_items(jid, node, self._channel_altinfo, args=(name,))
            return
        # Continue with the next
        self._fetch_source = None
        self._query_visible()

    def _update_error(self, iter_: Gtk.TreeIter, jid: str, node: str) -> None:
        # Switch to alternate query mode
        self.cache.get_items(jid, node, self._channel_altinfo)


class DiscussionGroupsBrowser(AgentBrowser):
    """
    For browsing pubsub-based discussion groups service
    """

    def __init__(self, account: str, jid: str, node: str) -> None:
        AgentBrowser.__init__(self, account, jid, node)

        # This will become set object when we get subscriptions; None means
        # we don't know yet which groups are subscribed
        self.subscriptions: set[str] | None = None

        # This will become our action widgets when we create them; None means
        # we don't have them yet (needed for check in callback)
        self.subscribe_button = None
        self.unsubscribe_button = None

        client = app.get_client(account)
        client.get_module("PubSub").get_subscriptions(
            jid, callback=self._on_subscriptions_received
        )

    def _create_treemodel(self):
        """
        Create treemodel for the window
        """
        # JID, node, name (with description) - pango markup,
        # don't have info?, subscribed?
        self.model = Gtk.TreeStore(str, str, str, bool, bool)
        # Sort by name
        self.model.set_sort_column_id(2, Gtk.SortType.ASCENDING)
        self.window.services_treeview.set_model(self.model)

        # Name column
        # Pango markup for name and description, description printed with
        # <small/> font
        renderer = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn(_("Name"))
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "markup", 2)
        col.set_resizable(True)
        self.window.services_treeview.insert_column(col, -1)
        self.window.services_treeview.set_headers_visible(True)

        # Subscription state
        renderer = Gtk.CellRendererToggle()
        col = Gtk.TreeViewColumn(_("Subscribed"))
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "inconsistent", 3)
        col.add_attribute(renderer, "active", 4)
        col.set_resizable(False)
        self.window.services_treeview.insert_column(col, -1)

        # Node Column
        renderer = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn(_("Node"))
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "markup", 1)
        col.set_resizable(True)
        self.window.services_treeview.insert_column(col, -1)

    def _in_list_foreach(
        self, model: Gtk.TreeModel, path: Gtk.TreePath, iter_: Gtk.TreeIter, node: str
    ) -> bool:
        if model[path][1] == node:
            self.in_list = True
            return True
        return False

    def _in_list(self, node: str) -> bool:
        self.in_list = False
        self.model.foreach(self._in_list_foreach, node)
        return self.in_list

    def _add_item(self, parent_node: str | None, item: DiscoItem, force: bool) -> None:
        """
        Called when we got basic information about new node from query. Show the
        item
        """
        node = item.node or ""
        name = item.name or ""
        jid = str(item.jid)

        if self.subscriptions is not None:
            dunno = False
            subscribed = node in self.subscriptions
        else:
            dunno = True
            subscribed = False

        name = GLib.markup_escape_text(name)
        name = "<b>%s</b>" % name

        if parent_node:
            parent_iter = self._get_iter(parent_node)
        else:
            parent_iter = None

        if not node or not self._in_list(node):
            self.model.append(parent_iter, [jid, node, name, dunno, subscribed])

    def _get_iter(self, node: str) -> Gtk.TreeIter | None:
        """Look for an iter with the given node"""
        self.found_iter: Gtk.TreeIter | None = None

        def is_node(
            model: Gtk.TreeModel, _path: Gtk.TreePath, iter_: Gtk.TreeIter, node: str
        ) -> bool:
            if model[iter_][1] == node:
                self.found_iter = iter_
                return True
            return False

        self.model.foreach(is_node, node)
        return self.found_iter

    def _add_actions(self) -> None:
        self.subscribe_button = Gtk.Button(label=_("_Subscribe"), use_underline=True)
        self.subscribe_button.set_sensitive(False)
        self.subscribe_button.connect("clicked", self._on_subscribe_button_clicked)
        self.window.action_buttonbox.append(self.subscribe_button)
        self.subscribe_button.set_visible(True)

        self.unsubscribe_button = Gtk.Button(
            label=_("_Unsubscribe"), use_underline=True
        )
        self.unsubscribe_button.set_sensitive(False)
        self.unsubscribe_button.connect("clicked", self._on_unsubscribe_button_clicked)
        self.window.action_buttonbox.append(self.unsubscribe_button)
        self.unsubscribe_button.set_visible(True)

    def _clean_actions(self) -> None:
        assert self.subscribe_button is not None
        assert self.unsubscribe_button is not None

        self.window.action_buttonbox.remove(self.subscribe_button)
        self.subscribe_button = None

        self.window.action_buttonbox.remove(self.unsubscribe_button)
        self.unsubscribe_button = None

    def update_actions(self) -> None:
        """
        Called when user selected a row. Make subscribe/unsubscribe buttons
        sensitive appropriately
        """
        if self.subscribe_button is None or self.unsubscribe_button is None:
            # _add_actions was not called yet
            return

        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if not iter_ or self.subscriptions is None:
            # No item selected or no subscriptions info, all buttons are
            # insensitive
            self.subscribe_button.set_sensitive(False)
            self.unsubscribe_button.set_sensitive(False)
            return

        subscribed = model.get_value(iter_, 4)  # 4 = subscribed?
        self.subscribe_button.set_sensitive(not subscribed)
        self.unsubscribe_button.set_sensitive(subscribed)

    def _on_subscribe_button_clicked(self, _button: Gtk.Button) -> None:
        """
        Called when 'subscribe' button is pressed. Send subscription request
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if iter_ is None:
            return

        node = model.get_value(iter_, 1)  # 1 = groupnode

        client = app.get_client(self.account)
        client.get_module("PubSub").subscribe(
            node, self.jid, callback=self._on_subscribe_response, user_data=node
        )

    def _on_unsubscribe_button_clicked(self, _button: Gtk.Button) -> None:
        """
        Called when 'unsubscribe' button is pressed. Send unsubscription request
        """
        model, iter_ = self.window.services_treeview.get_selection().get_selected()
        if iter_ is None:
            return

        node = model.get_value(iter_, 1)  # 1 = groupnode

        client = app.get_client(self.account)
        client.get_module("PubSub").unsubscribe(
            node, self.jid, callback=self._on_unsubscribe_response, user_data=node
        )

    def _on_subscriptions_received(self, task: Task) -> None:
        try:
            result = cast(list[PubSubSubscription], task.finish())
        except Exception as error:
            log.exception("Failed to retrieve subscriptions: %s", error)
            return

        groups = {sub.node for sub in result}
        if groups:
            log.info("Received subscribed nodes: %s", ", ".join(groups))

        self.subscriptions = groups

        # Try to setup existing items in model
        model = self.window.services_treeview.get_model()
        assert model is not None
        for row in model:
            # 1 = group node
            # 3 = insensitive checkbox for subscribed
            # 4 = subscribed?
            groupnode = row[1]
            row[3] = False
            row[4] = groupnode in groups

        # we now know subscriptions, update button states
        self.update_actions()

    def _on_subscribe_response(self, task: Task) -> None:
        groupnode = cast(str, task.get_user_data())
        try:
            result = cast(PubSubSubscription, task.finish())
        except Exception as error:
            log.exception("Failed to subscribe: %s, %s", groupnode, error)
            return

        if result.subscription != "subscribed":
            log.warning("Unhandled subscription value: %s", result.subscription)
            return

        assert self.subscriptions is not None
        self.subscriptions.add(groupnode)

        model = self.window.services_treeview.get_model()
        assert model is not None
        for row in model:
            if row[1] == groupnode:  # 1 = groupnode
                row[4] = True
                break

        self.update_actions()

    def _on_unsubscribe_response(self, task: Task) -> None:
        groupnode = cast(str, task.get_user_data())
        try:
            task.finish()
        except Exception as error:
            log.exception("Failed to unsubscribe: %s, %s", groupnode, error)
            return

        assert self.subscriptions is not None
        self.subscriptions.remove(groupnode)

        model = self.window.services_treeview.get_model()
        assert model is not None
        for row in model:
            if row[1] == groupnode:  # 1 = groupnode
                row[4] = False
                break

        self.update_actions()


# Fill the global agent type info dictionary
_agent_type_info = _gen_agent_type_info()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast

import hashlib
import logging
import sys
from pathlib import Path
from urllib.parse import urlparse

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common import configpaths
from gajim.common.const import ALL_MIME_TYPES
from gajim.common.enum import FTState
from gajim.common.enum import PreviewState
from gajim.common.file_transfer_manager import FileTransfer
from gajim.common.helpers import determine_proxy
from gajim.common.i18n import _
from gajim.common.multiprocess.http import CancelledError
from gajim.common.multiprocess.http import ContentTypeNotAllowed
from gajim.common.multiprocess.http import HTTPStatusError
from gajim.common.multiprocess.http import MaxContentLengthExceeded
from gajim.common.util.preview import contains_audio_streams
from gajim.common.util.preview import get_icon_for_mime_type
from gajim.common.util.preview import get_image_paths
from gajim.common.util.preview import get_size_and_mime_type
from gajim.common.util.preview import is_audio
from gajim.common.util.preview import is_image
from gajim.common.util.preview import is_video
from gajim.common.util.preview import UrlPreview

from gajim.gtk.menus import get_preview_menu
from gajim.gtk.preview.audio import AudioPreviewWidget
from gajim.gtk.preview.file_control_buttons import FileControlButtons
from gajim.gtk.preview.image import ImagePreviewWidget
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import GajimPopover

log = logging.getLogger("gajim.gtk.preview")


@Gtk.Template.from_string(string=get_ui_string("preview/preview.ui"))
class PreviewWidget(Gtk.Box, SignalManager):
    """Process Diagram

    flowchart TD

    %% Nodes
        A[Start]
        B{File Available}
        C[Downloading]
        D{Error?}
        E{Download Error?}
        F[Offer Download]
        G[Error]
        H[Downloaded]
        I{Previewable?}
        K[Display]
        L[Do Nothing]

    %% Defining the styles
        classDef Green fill:#50C878;

    %% Assigning styles to nodes
        class C,F,G,H,J,K Green;

    %% Links
        A --> B
        B -->|No| C
        B -->|Yes| H
        C --> D
        D --> |Yes| E
        E --> |No| F
        E --> |Yes| G
        F --> C
        D --> |No| H
        H --> I
        I --> |No| L
        I --> |Yes| K
    """

    __gtype_name__ = "PreviewWidget"

    _stack: Gtk.Stack = Gtk.Template.Child()
    _icon_button: Gtk.Button = Gtk.Template.Child()
    _mime_image: Gtk.Image = Gtk.Template.Child()
    _right_box: Gtk.Box = Gtk.Template.Child()
    _progress_box: Gtk.Box = Gtk.Template.Child()
    _progressbar: Gtk.ProgressBar = Gtk.Template.Child()
    _progress_text: Gtk.Label = Gtk.Template.Child()
    _cancel_download_button: Gtk.Button = Gtk.Template.Child()
    _content_box: Gtk.Box = Gtk.Template.Child()
    _link_button: Gtk.LinkButton = Gtk.Template.Child()
    _info_message_label: Gtk.Label = Gtk.Template.Child()
    _file_control_buttons: FileControlButtons = Gtk.Template.Child()
    _download_button: Gtk.Button = Gtk.Template.Child()

    def __init__(
        self, account: str, preview: UrlPreview, from_us: bool, context: str | None
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._from_us = from_us
        self._uri = preview.uri
        self._file_size = 0
        self._preview_id = hashlib.sha256(self._uri.encode()).hexdigest()
        self._preview_id_short = self._preview_id[:10]
        self._info_message = None
        self._http_obj = None
        self._state = PreviewState.INIT

        drag_source = Gtk.DragSource(actions=Gdk.DragAction.COPY)
        self._connect(drag_source, "prepare", self._on_drag_prepare)
        self._connect(drag_source, "drag-begin", self._on_drag_begin)
        self._connect(drag_source, "drag-end", self._on_drag_end)
        self.add_controller(drag_source)

        self._connect(
            self._cancel_download_button, "clicked", self._on_cancel_download_clicked
        )
        self._connect(self._download_button, "clicked", self._on_download_clicked)
        self._connect(self._icon_button, "clicked", self._on_icon_clicked)

        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        self._icon_button.set_cursor(pointer_cursor)
        self._cancel_download_button.set_cursor(pointer_cursor)
        self._download_button.set_cursor(pointer_cursor)

        self._menu_popover = GajimPopover(None)
        self.append(self._menu_popover)

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary_click, "pressed", self._on_preview_clicked)
        self.add_controller(gesture_secondary_click)

        # Set initial file attributes, guessed based on the uri
        self._filename = preview.file_name
        self._mime_type = preview.mime_type

        self._file_control_buttons.set_file_name(self._filename)
        self._mime_image.set_from_gicon(get_icon_for_mime_type(self._mime_type))

        self._orig_dir = configpaths.get("DOWNLOADS")
        self._thumb_dir = configpaths.get("DOWNLOADS_THUMB")

        self._urlparts = urlparse(self._uri)
        thumbnail_size = app.settings.get("preview_size")

        self._orig_path, self._thumb_path = get_image_paths(
            self._uri, self._urlparts, thumbnail_size, self._orig_dir, self._thumb_dir
        )

        self._link_button.set_uri(self._uri)
        self._link_button.set_tooltip_text(self._uri)
        self._link_button.set_label(self._uri)
        label = cast(Gtk.Label, self._link_button.get_child())
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_max_width_chars(32)

        if ftobj := app.ftm.get_transfer(self._preview_id):
            log.info(
                "Bind to existing transfer: %s %s", self._preview_id_short, self._uri
            )
            self._connect_to_ftobj(ftobj)
            return

        if self._orig_path.exists():
            self._mime_type, self._file_size = get_size_and_mime_type(self._orig_path)
            self._set_widget_state(PreviewState.DOWNLOADED)
            self._set_widget_state(PreviewState.DISPLAY)
            return

        max_content_length = app.settings.get("preview_max_file_size")
        if max_content_length > 0 and self._should_auto_preview(context):
            self._download_content(max_content_length, ALL_MIME_TYPES)

        else:
            self._info_message = _("Automatic preview disabled")
            self._set_widget_state(PreviewState.OFFER_DOWNLOAD)

    def do_unroot(self) -> None:
        self._disconnect_all()
        del self._menu_popover
        del self._http_obj
        Gtk.Box.do_unroot(self)

    def get_text(self) -> str:
        return self._uri

    def get_orig_path(self) -> Path:
        return self._orig_path

    def _on_drag_prepare(
        self, _drag_source: Gtk.DragSource, x: float, y: float
    ) -> Gdk.ContentProvider | None:
        if self._state not in (PreviewState.DOWNLOADED, PreviewState.DISPLAY):
            return None

        value = GObject.Value()
        value.init(PreviewWidget)
        value.set_object(self)

        return Gdk.ContentProvider.new_for_value(value)

    def _on_drag_begin(self, _drag_source: Gtk.DragSource, drag: Gdk.Drag) -> None:
        Gtk.DragIcon.set_from_paintable(
            drag,
            Gtk.WidgetPaintable().new(self),
            hot_x=0,
            hot_y=0,
        )

        app.window.highlight_dnd_targets(self, True)

    def _on_drag_end(
        self, _drag_source: Gtk.DragSource, _drag: Gdk.Drag, _delete_data: bool
    ) -> None:
        app.window.highlight_dnd_targets(self, False)

    def _should_auto_preview(self, context: str | None) -> bool:
        if self._from_us:
            return True

        return context != "public" or app.settings.get("preview_anonymous_muc")

    def _reset_progress(self) -> None:
        self._progress_text.set_label("0 %")
        self._progressbar.set_fraction(0)

    def _set_display_widget(self, widget: Gtk.Widget) -> None:
        self._stack.add_named(widget, "widget")
        self._stack.set_visible_child_name("widget")

    def _set_widget_state(self, state: PreviewState) -> None:
        log.info("Set widget state %s %s", self._preview_id_short, state.name)
        self._state = state

        # Handle PreviewState.DISPLAY state first, so if no fitting widget is found
        # all other widget states remain as in PreviewState.DOWNLOADED
        if state == PreviewState.DISPLAY:
            widget = None
            if is_video(self._mime_type) and sys.platform == "darwin":
                # https://dev.gajim.org/gajim/gajim/-/issues/12625
                pass

            elif is_image(self._mime_type) or is_video(self._mime_type):
                assert self._mime_type is not None
                widget = ImagePreviewWidget(
                    self._filename,
                    self._file_size,
                    self._mime_type,
                    self._orig_path,
                    self._thumb_path,
                )

            elif is_audio(self._mime_type):
                if (
                    app.audio_player is not None
                    and app.is_installed("GST")
                    and contains_audio_streams(self._orig_path)
                ):
                    widget = AudioPreviewWidget(
                        app.audio_player,
                        self._filename,
                        self._file_size,
                        self._orig_path,
                    )

            if widget is not None:
                self._connect(widget, "display-error", self._on_display_error)
                self._set_display_widget(widget)

            return

        self._download_button.set_visible(state == PreviewState.OFFER_DOWNLOAD)
        self._download_button.set_sensitive(state == PreviewState.OFFER_DOWNLOAD)

        self._file_control_buttons.set_file_name(self._filename)
        self._file_control_buttons.set_file_size(self._file_size)

        self._cancel_download_button.set_visible(state == PreviewState.DOWNLOADING)
        self._cancel_download_button.set_sensitive(state == PreviewState.DOWNLOADING)

        self._progress_box.set_visible(state == PreviewState.DOWNLOADING)

        self._mime_image.set_from_gicon(get_icon_for_mime_type(self._mime_type))
        self._link_button.set_visible(state == PreviewState.ERROR)

        self._info_message_label.set_text(self._info_message or "")
        self._info_message_label.set_tooltip_text(self._info_message or "")
        self._info_message_label.set_visible(self._info_message is not None)

        if state == PreviewState.DOWNLOADING:
            self._reset_progress()

        elif state == PreviewState.DOWNLOADED:
            self._file_control_buttons.set_path(self._orig_path)

    def _on_display_error(self, _widget: Any) -> None:
        self._stack.set_visible_child_name("preview")

    def _on_download_clicked(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        self._download_content()

    def _on_icon_clicked(self, button: Gtk.Button) -> None:
        if self._state == PreviewState.OFFER_DOWNLOAD:
            self._download_button.emit("clicked")

        elif self._state == PreviewState.DISPLAY:
            app.app.activate_action(
                "open-file", GLib.Variant("s", str(self._orig_path))
            )

    def _on_cancel_download_clicked(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        assert self._http_obj is not None
        self._http_obj.cancel()

    def _on_preview_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        encrypted = self._uri.startswith("aesgcm://")
        menu = get_preview_menu(self._uri, encrypted=encrypted)
        self._menu_popover.set_menu_model(menu)
        self._menu_popover.set_pointing_to_coord(x, y)
        self._menu_popover.popup()

    def _download_content(
        self,
        max_content_length: int = -1,
        allowed_content_types: set[str] | None = None,
    ) -> None:
        log.info("Start downloading: %s %s", self._preview_id_short, self._uri)

        obj = app.ftm.http_request(
            "GET",
            self._uri,
            self._preview_id,
            output=self._orig_path,
            with_progress=True,
            max_content_length=max_content_length,
            allowed_content_types=allowed_content_types,
            proxy=determine_proxy(self._account),
        )

        if obj is None:
            return

        self._connect_to_ftobj(obj)

    def _connect_to_ftobj(self, obj: FileTransfer) -> None:
        self._connect(obj, "finished", self._on_download_finished)
        self._connect(obj, "notify::progress", self._on_download_progress)
        self._http_obj = obj
        self._info_message = None

        if obj.state <= FTState.IN_PROGRESS:
            self._set_widget_state(PreviewState.DOWNLOADING)

    def _on_download_progress(
        self, ftobj: FileTransfer, _param: GObject.ParamSpec
    ) -> None:
        progress = ftobj.get_property("progress")
        total = ftobj.get_total()
        if total is None:
            self._progressbar.set_pulse_step(0.1)
            self._progressbar.pulse()
            return

        fraction = progress / total
        self._progress_text.set_label(f"{int(fraction * 100)} %")
        self._progressbar.set_fraction(fraction)

    def _on_download_finished(
        self,
        ftobj: FileTransfer,
    ) -> None:
        self._disconnect_object(ftobj)
        assert self._orig_path is not None
        self._info_message = None
        next_state = PreviewState.DOWNLOADED

        try:
            ftobj.raise_for_error()
        except ContentTypeNotAllowed as error:
            log.info("Not an allowed content type: %s, %s", error, self._uri)
            next_state = PreviewState.OFFER_DOWNLOAD

        except MaxContentLengthExceeded as error:
            log.info('File size (%s) too big for URL: "%s"', error, self._uri)
            self._info_message = _("Automatic preview disabled (file too big)")
            next_state = PreviewState.OFFER_DOWNLOAD

        except HTTPStatusError as error:
            log.info("Status error for %s: %s", self._uri, error)
            self._info_message = str(error)
            next_state = PreviewState.ERROR

        except OverflowError as error:
            log.info("Content-Length overflow for %s: %s", self._uri, error)
            next_state = PreviewState.ERROR

        except CancelledError:
            log.info("Download cancelled for %s", self._uri)
            next_state = PreviewState.OFFER_DOWNLOAD

        except Exception:
            log.exception("Unknown error for: %s", self._uri)
            self._info_message = _("Unknown Error")
            next_state = PreviewState.ERROR

        finally:
            if metadata := ftobj.get_metadata():
                if self._orig_path.exists():
                    self._mime_type, self._file_size = get_size_and_mime_type(
                        self._orig_path
                    )

                else:
                    self._mime_type = metadata.content_type or ""
                    self._file_size = metadata.content_length or -1

        self._set_widget_state(next_state)

        if ftobj.state != FTState.FINISHED:
            return

        log.info("File stored: %s %s", self._preview_id_short, self._orig_path.name)

        self._set_widget_state(PreviewState.DISPLAY)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import NoReturn

import argparse
import io
import operator
import shlex
import sys
from collections.abc import Callable

from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import regex
from gajim.common.helpers import Observable
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.util.muc import is_affiliation_change_allowed
from gajim.common.util.muc import is_role_change_allowed


def split_argument_string(string: str) -> list[str]:
    '''
    Split a string with shlex.split
    '''

    lex = shlex.shlex(string, posix=True)
    lex.whitespace_split = True
    lex.commenters = ''
    result: list[str] = []

    try:
        for token in lex:
            result.append(token)
    except ValueError:
        # If end-of-string is reached and there is a invalid state
        # ValueError is raised. Still add the partial token to the result.
        result.append(lex.token)

    return result


def get_usage_from_command(cmd: argparse.ArgumentParser,
                           command_name: str
                           ) -> str:

    with io.StringIO() as output:
        cmd.print_usage(file=output)
        usage = output.getvalue()

    usage = usage.split(f'ChatCommands {command_name}')[1]
    return usage.strip()


class ArgumentParserError(Exception):
    pass


class CommandError(Exception):
    pass


class CommandFailed(Exception):
    pass


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message: str) -> NoReturn:
        raise ArgumentParserError(message)


class ChatCommands(Observable):
    def __init__(self) -> None:
        Observable.__init__(self)

    def init(self) -> None:
        if sys.version_info >= (3, 14):
            # Disable colors because we depend on a stable format for the
            # usage string
            self._parser = ArgumentParser(prog='ChatCommands', color=False)
        else:
            self._parser = ArgumentParser(prog='ChatCommands')
        self._sub_parser = self._parser.add_subparsers(title='Commands')
        self._commands: dict[str, tuple[list[str], str]] = {}

        self._create_commands()

    def get_commands(self, used_in: str) -> list[tuple[str, str]]:
        commands: list[tuple[str, str]] = []
        for cmd_name, cmd in self._commands.items():
            if used_in in cmd[0]:
                commands.append((cmd_name, cmd[1]))
        commands.sort(key=operator.itemgetter(0))
        return commands

    def _generate_help(self, used_in: str) -> str:
        help_text = 'Commands:\n\n'
        commands = self.get_commands(used_in)
        for command in commands:
            name, usage = command
            if name == 'help':
                continue
            help_text += f'  {name:<15} {usage}\n'
        return help_text

    def make_parser(self,
                    command_name: str,
                    callback: Callable[..., Any],
                    **kwargs: Any) -> ArgumentParser:

        '''Add and return a subparser and initialize
        it with the command name.
        '''

        parser = self._sub_parser.add_parser(command_name,
                                             add_help=False,
                                             **kwargs)
        parser.set_defaults(command_name=command_name,
                            exec=callback)
        return parser

    def add_command(self,
                    command_name: str,
                    used_in: list[str],
                    cmd: argparse.ArgumentParser
                    ) -> None:

        usage = get_usage_from_command(cmd, command_name)
        self._commands[command_name] = (used_in, usage)

    def parse(self, type_: str, arg_string: str) -> bool:
        if regex.COMMAND_REGEX.match(arg_string) is None:
            return False

        arg_list = split_argument_string(arg_string[1:])

        command_name = arg_list[0]
        command = self._commands.get(command_name)
        if command is None or type_ not in command[0]:
            unknown_command = _('Unknown command: %s') % command_name
            hint = _('Use "//" to start a message with a slash "/".')
            self.notify('command-not-found',
                        f'{unknown_command}\n{hint}')
            raise CommandFailed

        try:
            args = self._parser.parse_args(arg_list)
        except ArgumentParserError as error:
            self.notify('command-error', str(error))
            raise CommandFailed

        args.used_in = type_

        try:
            result = args.exec(args)
        except CommandError as error:
            self.notify('command-error', str(error))
            raise CommandFailed

        if result == 'silent':
            return True

        if result is None:
            result = _('Command executed successfully')

        self.notify('command-result', result)
        return True

    def _create_commands(self) -> None:
        parser = self.make_parser('help', self._help_command)
        self.add_command('help', ['chat', 'groupchat', 'pm'], parser)

        parser = self.make_parser('status', self._status_command)
        parser.add_argument('status',
                            choices=['online', 'away', 'xa', 'dnd'])
        parser.add_argument('message', default=None, nargs='?')
        self.add_command('status', ['chat', 'groupchat', 'pm'], parser)

        parser = self.make_parser('invite', self._invite_command)
        parser.add_argument('address')
        parser.add_argument('reason', default=None, nargs='?')
        self.add_command('invite', ['groupchat'], parser)

        parser = self.make_parser('ban', self._ban_command)
        parser.add_argument('who')
        parser.add_argument('reason', default=None, nargs='?')
        self.add_command('ban', ['groupchat'], parser)

        parser = self.make_parser('affiliate', self._affiliate_command)
        parser.add_argument('who')
        parser.add_argument('affiliation',
                            choices=['owner', 'admin', 'member', 'none'])
        self.add_command('affiliate', ['groupchat'], parser)

        parser = self.make_parser('kick', self._kick_command)
        parser.add_argument('who')
        parser.add_argument('reason', default=None, nargs='?')
        self.add_command('kick', ['groupchat'], parser)

        parser = self.make_parser('role', self._role_command)
        parser.add_argument('who')
        parser.add_argument('role',
                            choices=['moderator', 'participant', 'visitor'])
        self.add_command('role', ['groupchat'], parser)

        parser = self.make_parser('me', self._me_command)
        parser.add_argument('message', default=None, nargs='*')
        self.add_command('me', ['chat', 'groupchat', 'pm'], parser)

        app.plugin_manager.extension_point('add_commands', self)

    def _help_command(self, args: Any) -> str:
        return self._generate_help(args.used_in)

    def _status_command(self, args: Any) -> None:
        for client in app.get_clients():
            if not app.settings.get_account_setting(client.account,
                                                    'sync_with_global_status'):
                continue

            if not client.state.is_available:
                continue

            client.change_status(args.status,
                                 args.message or client.status_message)

    def _check_if_joined(self) -> GroupchatContact:
        contact = app.window.get_control().contact
        assert isinstance(contact, GroupchatContact)
        if not contact.is_joined:
            raise CommandError(_('You are currently not '
                                 'joined this group chat'))
        return contact

    def _invite_command(self, args: Any) -> None:
        contact = self._check_if_joined()

        try:
            jid = JID.from_string(args.address)
        except Exception:
            raise CommandError(_('Invalid address: %s') % args.address)

        if jid.is_full or jid.localpart is None:
            raise CommandError(_('Invalid address: %s') % args.address)

        client = app.get_client(contact.account)
        client.get_module('MUC').invite(contact.jid, args.address, args.reason)

    def _change_affiliation(self,
                            nick_or_address: str,
                            affiliation: str,
                            reason: str | None) -> None:

        contact = self._check_if_joined()

        nick_list = contact.get_user_nicknames()
        if nick_or_address in nick_list:
            participant = contact.get_resource(nick_or_address)
            self_contact = contact.get_self()
            assert self_contact is not None
            if not is_affiliation_change_allowed(self_contact,
                                                 participant,
                                                 'outcast'):
                raise CommandError(_('You have insufficient permissions'))

            jid = participant.real_jid

        else:
            try:
                jid = JID.from_string(nick_or_address)
            except Exception:
                raise CommandError(_('Invalid address: %s') % nick_or_address)

            if jid.is_full or jid.localpart is None:
                raise CommandError(_('Invalid address: %s') % nick_or_address)

        client = app.get_client(contact.account)
        client.get_module('MUC').set_affiliation(
            contact.jid,
            {jid: {'affiliation': affiliation,
                   'reason': reason}})

    def _ban_command(self, args: Any) -> None:
        self._change_affiliation(args.who, 'outcast', args.reason)

    def _affiliate_command(self, args: Any) -> None:
        self._change_affiliation(args.who, args.affiliation, None)

    def _change_role(self, nick: str, role: str, reason: str | None) -> None:

        contact = self._check_if_joined()

        nick_list = contact.get_user_nicknames()
        if nick not in nick_list:
            raise CommandError(_('User %s not found') % nick)

        participant = contact.get_resource(nick)
        self_contact = contact.get_self()
        assert self_contact is not None
        if not is_role_change_allowed(self_contact, participant):
            raise CommandError(_('You have insufficient permissions'))

        client = app.get_client(contact.account)
        client.get_module('MUC').set_role(contact.jid,
                                          nick,
                                          role,
                                          reason)

    def _kick_command(self, args: Any) -> None:
        self._change_role(args.who, 'none', args.reason)

    def _role_command(self, args: Any) -> None:
        self._change_role(args.who, args.role, None)

    def _me_command(self, _args: Any) -> str:
        app.window.activate_action('win.send-message', None)
        return 'silent'



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any

import logging
import sys
from pathlib import Path

from gi.repository import Gio
from gi.repository import GLib

from gajim.common import app

if sys.platform == 'win32' or typing.TYPE_CHECKING:
    import winsound

if sys.platform == 'darwin' or typing.TYPE_CHECKING:
    from AppKit import NSSound

log = logging.getLogger('gajim.c.sound')


class PlaySound:

    def play(self, path: Path, loop: bool = False) -> None:
        raise NotImplementedError

    def stop(self) -> None:
        raise NotImplementedError

    def loop_in_progress(self) -> bool:
        raise NotImplementedError


class PlatformWindows(PlaySound):

    def __init__(self) -> None:
        self._loop_in_progress = False

    def play(self, path: Path, loop: bool = False) -> None:
        if self._loop_in_progress:
            return

        assert winsound is not None
        flags = (winsound.SND_FILENAME |
                 winsound.SND_ASYNC |
                 winsound.SND_NODEFAULT)
        if loop:
            self._loop_in_progress = True
            flags = flags | winsound.SND_LOOP

        try:
            winsound.PlaySound(str(path), flags)
        except Exception:
            log.exception('Sound Playback Error')

    def stop(self) -> None:
        assert winsound is not None
        try:
            winsound.PlaySound(None, 0)
        except Exception:
            log.exception('Sound Playback Error')
        self._loop_in_progress = False

    def loop_in_progress(self) -> bool:
        return self._loop_in_progress


class PlatformMacOS(PlaySound):

    def play(self, path: Path, loop: bool = False) -> None:
        assert NSSound is not None
        sound = NSSound.alloc()
        sound.initWithContentsOfFile_byReference_(str(path), True)
        sound.play()

    def stop(self) -> None:
        pass

    def loop_in_progress(self) -> bool:
        return False


class PlatformUnix(PlaySound):

    def __init__(self) -> None:
        self._cancellable: Gio.Cancellable | None = None

    def play(self, path: Path, loop: bool = False) -> None:
        if not app.is_installed('GSOUND'):
            return

        if self.loop_in_progress():
            return

        attrs = {'media.filename': str(path)}

        if loop:
            self._play_loop(attrs)
        else:
            try:
                app.gsound_ctx.play_simple(attrs, self._cancellable)
            except GLib.Error as error:
                log.error('Could not play sound: %s', error.message)
                self._cancellable = None

    def _play_loop(self, attrs: dict[str, str]) -> None:
        self._cancellable = Gio.Cancellable()
        try:
            app.gsound_ctx.play_full(attrs,
                                     self._cancellable,
                                     self._on_play_finished,
                                     attrs)
        except GLib.Error as error:
            log.error('Could not play sound: %s', error.message)
            self._cancellable = None

    def _on_play_finished(self,
                          _context: Any,
                          res: Gio.AsyncResult,
                          user_data: dict[str, str]) -> None:
        try:
            app.gsound_ctx.play_full_finish(res)
        except GLib.Error as error:
            quark = GLib.quark_try_string('g-io-error-quark')
            if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                self._cancellable = None
                return
            log.error('Could not play sound: %s', error.message)

        self._play_loop(user_data)

    def stop(self) -> None:
        if self._cancellable is None or self._cancellable.is_cancelled():
            return
        self._cancellable.cancel()
        self._cancellable = None

    def loop_in_progress(self) -> bool:
        return self._cancellable is not None


def _init_platform() -> PlaySound:
    if sys.platform == 'win32':
        return PlatformWindows()

    if sys.platform == 'darwin':
        return PlatformMacOS()

    return PlatformUnix()


_platform_player = _init_platform()


def play(path: Path, loop: bool = False) -> None:
    _platform_player.play(path, loop)


def stop():
    _platform_player.stop()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0191: Blocking Command

from __future__ import annotations

from typing import Any

from collections.abc import Generator
from collections.abc import Iterator

import nbxmpp
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.structs import BlockingProperties
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import StanzaHandler
from nbxmpp.types import BlockingReportValues

from gajim.common import app
from gajim.common import types
from gajim.common.events import ContactBlocked
from gajim.common.events import FeatureDiscovered
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.util import as_task


class Blocking(BaseModule):

    _nbxmpp_extends = 'Blocking'
    _nbxmpp_methods = [
        'block',
        'unblock',
        'request_blocking_list',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.blocked: set[JID] = set()

        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._blocking_push_received,
                          typ='set',
                          ns=Namespace.BLOCKING),
        ]

        self.supported = False

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.BLOCKING not in info.features:
            return

        self.supported = True
        app.ged.raise_event(
            FeatureDiscovered(account=self._account,
                              feature=Namespace.BLOCKING))

        self._log.info('Discovered blocking: %s', info.jid)

    def is_blocked(self, jid: JID) -> bool:
        return jid in self.blocked

    def block(self, jids: list[JID], report: BlockingReportValues | None) -> None:
        self._nbxmpp("Blocking").block(jids, report)
        for jid in jids:
            app.ged.raise_event(ContactBlocked(account=self._account, jid=jid))

    @as_task
    def get_blocking_list(
        self
    ) -> Generator[Any, set[JID]]:
        _task = yield  # noqa: F841

        blocking_list = yield self._nbxmpp('Blocking').request_blocking_list()

        raise_if_error(blocking_list)

        self.blocked = blocking_list
        for contact in self._get_contacts_from_jids(blocking_list):
            contact.set_blocked()

        yield blocking_list

    @as_task
    def update_blocking_list(self,
                             block: set[JID],
                             unblock: set[JID]
                             ) -> Generator[Any, bool]:
        _task = yield  # noqa: F841

        if block:
            result = yield self._nbxmpp("Blocking").block(list(block))
            raise_if_error(result)

        if unblock:
            result = yield self.unblock(unblock)
            raise_if_error(result)

        yield True

    def _blocking_push_received(self,
                                _con: types.NBXMPPClient,
                                _stanza: Iq,
                                properties: BlockingProperties
                                ) -> None:
        if not properties.is_blocking:
            return

        if properties.blocking.unblock_all:
            unblock = set(self.blocked)
            self.blocked = set()
            self._unblock_jids(unblock)
            raise nbxmpp.NodeProcessed

        if properties.blocking.unblock:
            self.blocked -= properties.blocking.unblock
            self._unblock_jids(properties.blocking.unblock)

        if properties.blocking.block:
            block = properties.blocking.block
            self.blocked.update(block)
            for contact in self._get_contacts_from_jids(block):
                self._log.info('Block Push: %s', contact.jid)
                contact.set_blocked()

        raise nbxmpp.NodeProcessed

    def _unblock_jids(self, jids: set[JID]) -> None:
        for contact in self._get_contacts_from_jids(jids):
            contact.set_unblocked()
            self._presence_probe(contact.jid)
            self._log.info('Unblock Push: %s', contact.jid)

    def _get_contacts_from_jids(self,
                                jids: set[JID]
                                ) -> Iterator[BareContact]:

        module = self._con.get_module('Contacts')

        for jid in jids:
            if jid.resource is not None:
                # Currently not supported by GUI
                continue

            if jid.is_domain:
                contacts = module.get_contacts_with_domain(jid.domain)
                for contact in contacts:

                    if not isinstance(contact, BareContact):
                        continue

                    yield contact

                continue

            contact = module.get_contact_if_exists(jid)
            if not isinstance(contact, BareContact):
                continue

            yield contact

    def _presence_probe(self, jid: JID) -> None:
        self._log.info('Presence probe: %s', jid)
        # Send a presence Probe to get the current Status
        probe = nbxmpp.Presence(jid, 'probe', frm=self._con.get_own_jid())
        self._nbxmpp().send(probe)
